import logging
import os
from collections import defaultdict
from typing import Sequence, Tuple

import numpy as np
from keras import Input

from medsegpy import config
from medsegpy.modeling import Model, get_model, model_from_json
from medsegpy.modeling.layers.convolutional import ConvStandardized2D
from medsegpy.modeling.layers.normalization import GroupNormalization
from medsegpy.modeling.meta_arch import build_model
from medsegpy.modeling.meta_arch.build import META_ARCH_REGISTRY
from medsegpy.utils import dl_utils

logger = logging.getLogger(__name__)


def get_all_layers(model: Model) -> Sequence[Tuple]:
    name_to_layer = []
    for layer in model.layers:
        if "input" not in layer.name:
            if "model" in layer.name:
                name_to_layer.extend(get_all_layers(layer))
            else:
                name_to_layer.append((layer.name, layer))
    return name_to_layer


def word_contains_words(
    word: str, include_words: Sequence[str], exclude_words: Sequence[str]
) -> bool:
    for iw in include_words:
        if iw not in word:
            return False
    for ew in exclude_words:
        if ew in word:
            return False
    return True


def remove_keys_from_config(config: dict):
    keys = ["name", "trainable", "kernel_initializer", "bias_initializer"]
    for key in keys:
        _ = config.pop(key, None)


def get_layers_to_load(model: Model, cfg: config.Config) -> Tuple[Sequence, Sequence]:
    # Get instance of config used to train the pretrained model
    assert type(cfg.PRETRAINED_WEIGHTS_PATH) is dict, "'PRETRAINED_WEIGHTS_PATH' must a dictionary"
    assert os.path.exists(cfg.PRETRAINED_CONFIG_PATH), (
        f"'PRETRAINED_CONFIG_PATH' does not exist! " f"Path: {cfg.PRETRAINED_CONFIG_PATH}"
    )
    pt_info = cfg.PRETRAINED_WEIGHTS_PATH
    pt_config_path = cfg.PRETRAINED_CONFIG_PATH
    pt_model_name = config.get_model_name(pt_config_path)
    pt_config = config.get_config(pt_model_name, create_dirs=False)
    pt_config.merge_from_file(pt_config_path)

    # Get path of weights to load
    weight_file = pt_info["path"]
    if not weight_file:
        # If weight_file not given, choose best weights based on
        # performance on validation set
        weight_file = dl_utils.get_weights(pt_config.OUTPUT_DIR)
    logger.info(f"Loading specific weights from {weight_file}")

    pt_model = None
    pt_model_json_file = os.path.join(pt_config.OUTPUT_DIR, "model.json")
    # Try loading pretrained model from JSON file, if this training run
    # is not the frozen run of self-supervised learning
    if os.path.isfile(pt_model_json_file):
        try:
            with open(pt_model_json_file) as f:
                json_str = f.read()
            pt_model = model_from_json(
                json_str,
                custom_objects={
                    "Model": Model,
                    "ConvStandardized2D": ConvStandardized2D,
                    "GroupNormalization": GroupNormalization,
                },
            )
        except Exception as e:
            print(e)

    if pt_model is None:
        # Build pretrained model from config
        pt_builder = META_ARCH_REGISTRY.get(pt_model_name)(pt_config)
        pt_model = pt_builder.build_model()
        logger.info("Built pretrained model from config")
    else:
        logger.info("Built pretrained model from JSON")

    # Load weights into pretrained model
    pt_model.load_weights(weight_file)

    # Find all layers of both pretrained and current model
    pt_layer_list = get_all_layers(pt_model)
    cur_layer_list = get_all_layers(model)
    pt_idx_layers = defaultdict(list)
    cur_idx_layers = defaultdict(list)
    weights_to_load = pt_info["weights"]

    # Determine list of layers to load
    assert type(weights_to_load) is list, "Key 'weights' of PRETRAINED_WEIGHTS_PATH must be a list"
    pt_layer_name_to_idx = {}
    cur_layer_name_to_idx = {}
    total_pt_load_layers = []
    total_cur_load_layers = []
    for weight_idx, weight_info in enumerate(weights_to_load):
        assert "include_words" in weight_info, (
            f"Key 'include_word' is required in " f"weight dictionary at index {weight_idx}"
        )
        assert "exclude_words" in weight_info, (
            f"Key 'exclude_word' is required in " f"weight dictionary at index {weight_idx}"
        )
        assert "slice_indices" in weight_info, (
            f"Key 'slice_indices' is required in " f"weight dictionary at index {weight_idx}"
        )

        # Check format of range of weight indices for each weight key
        range_list = weight_info["slice_indices"]
        assert len(range_list) == 2, (
            f"The length of the list for key 'slice_indices' must "
            f"2. Error occurs for weight dictionary at index {weight_idx}."
        )
        assert isinstance(range_list[0], int), (
            f"The first value in the list for key 'slice_indices' must "
            f"be an integer. Error occurs for weight dictionary at "
            f"index {weight_idx}."
        )

        for pt_layer_tuple in pt_layer_list:
            layer_name = pt_layer_tuple[0]
            if word_contains_words(
                layer_name, weight_info["include_words"], weight_info["exclude_words"]
            ):
                assert layer_name not in pt_layer_name_to_idx, (
                    f"Layer {layer_name} of pretrained model is already "
                    f"assigned to weight dictionary at index "
                    f"{pt_layer_name_to_idx[layer_name]}"
                )
                pt_idx_layers[weight_idx].append(pt_layer_tuple)
                pt_layer_name_to_idx[layer_name] = weight_idx
        assert (
            weight_idx in pt_idx_layers
        ), f"No layers found for weight dictionary at index {weight_idx}"
        for cur_layer_tuple in cur_layer_list:
            layer_name = cur_layer_tuple[0]
            if word_contains_words(
                layer_name, weight_info["include_words"], weight_info["exclude_words"]
            ):
                assert layer_name not in cur_layer_name_to_idx, (
                    f"Layer {layer_name} of current model is already "
                    f"assigned to weight dictionary at index "
                    f"{cur_layer_name_to_idx[layer_name]}"
                )
                cur_idx_layers[weight_idx].append(cur_layer_tuple)
                cur_layer_name_to_idx[layer_name] = weight_idx
                if range_list[1] != "until":
                    matched_pt_layer_tuple = pt_idx_layers[weight_idx][
                        len(cur_idx_layers[weight_idx]) - 1
                    ]
                    matched_pt_config = matched_pt_layer_tuple[1].get_config()
                    cur_config = cur_layer_tuple[1].get_config()
                    remove_keys_from_config(matched_pt_config)
                    remove_keys_from_config(cur_config)
                    assert matched_pt_config == cur_config, (
                        f"Config of layer in current model with name "
                        f"{layer_name} does not match config of corresponding "
                        f"layer in pretrained model with name "
                        f"{matched_pt_layer_tuple[0]}"
                    )
        assert (
            weight_idx in cur_idx_layers
        ), f"No layers found for weight dictionary at index {weight_idx}"

        # Check that both the current model and the pretrained model have
        # the same number of layers in each key group
        num_pt_layers = len(pt_idx_layers[weight_idx])
        num_cur_layers = len(cur_idx_layers[weight_idx])
        assert num_pt_layers == num_cur_layers, (
            f"Pretrained model and current model have different numbers "
            f"of layers for weight dictionary at index {weight_idx}. "
            f"Pretrained model has {num_pt_layers} and current model "
            f"has {num_cur_layers}."
        )

        if range_list[1] == "until":
            pt_load_layers = []
            cur_load_layers = []
            until_cur_layer_tuple = cur_idx_layers[weight_idx][0]
            for layer_idx in range(len(cur_layer_list)):
                cur_layer_tuple = cur_layer_list[layer_idx]
                pt_layer_tuple = pt_layer_list[layer_idx]
                pt_layer_name = pt_layer_tuple[0]
                cur_layer_name = cur_layer_tuple[0]

                # Check if current layer is the same as pretrained layer
                if not cur_layer_name == until_cur_layer_tuple[0]:
                    cur_config = cur_layer_tuple[1].get_config()
                    matched_pt_config = pt_layer_tuple[1].get_config()
                    remove_keys_from_config(cur_config)
                    remove_keys_from_config(matched_pt_config)
                    assert cur_config == matched_pt_config, (
                        f"Config of layer in current model with name "
                        f"{cur_layer_tuple[0]} does not match config of "
                        f"corresponding layer in pretrained model with name "
                        f"{pt_layer_tuple[0]}"
                    )
                    assert pt_layer_name not in pt_layer_name_to_idx, (
                        f"Layer {pt_layer_name} of pretrained model is already "
                        f"assigned to weight dictionary at index "
                        f"{pt_layer_name_to_idx[pt_layer_name]}"
                    )
                    assert cur_layer_name not in cur_layer_name_to_idx, (
                        f"Layer {cur_layer_name} of current model is already "
                        f"assigned to weight dictionary at index "
                        f"{cur_layer_name_to_idx[cur_layer_name]}"
                    )
                    pt_load_layers.append(pt_layer_tuple)
                    pt_layer_name_to_idx[pt_layer_name] = weight_idx
                    cur_load_layers.append(cur_layer_tuple)
                    cur_layer_name_to_idx[cur_layer_name] = weight_idx
                else:
                    break
        else:
            if range_list[1] == "None":
                range_list[1] = None
            assert range_list[1] is None or isinstance(range_list[1], int), (
                f"The second value in the list for key 'slice_indices' must "
                f"be either None or an integer. Error occurs for weight "
                f"dictionary at index {weight_idx}."
            )

            pt_load_layers = pt_idx_layers[weight_idx][range_list[0] : range_list[1]]
            cur_load_layers = cur_idx_layers[weight_idx][range_list[0] : range_list[1]]
        total_pt_load_layers.extend(pt_load_layers)
        total_cur_load_layers.extend(cur_load_layers)

    assert len(total_pt_load_layers) == len(total_cur_load_layers), (
        f"Must have same number of layers in pretrain layer list and "
        f"current layer list. Found {len(total_pt_load_layers)} pretrain "
        f"layers and {len(total_cur_load_layers)} current layers."
    )

    return total_pt_load_layers, total_cur_load_layers


def load_specific_weights(model: Model, cfg: config.Config, debug: bool = True):
    pt_load_layers, cur_load_layers = get_layers_to_load(model, cfg)
    num_layers_loaded = 0
    loaded_layer_names = []
    for cur_layer_tuple, pt_layer_tuple in zip(cur_load_layers, pt_load_layers):
        cur_layer_name = cur_layer_tuple[0]
        pt_layer = pt_layer_tuple[1]
        cur_layer = cur_layer_tuple[1]
        layer_weight = pt_layer.get_weights()
        if layer_weight:
            prev_weight = cur_layer.get_weights()
            cur_layer.set_weights(layer_weight)
            new_weight = cur_layer.get_weights()
            assert np.any(new_weight[0] != prev_weight[0]), (
                f"Loading weights for layer {cur_layer_name} "
                f"in current model did not work: Weights are still "
                f"random!"
            )
            assert np.all(new_weight[0] == layer_weight[0]), (
                f"Loading weights for layer {cur_layer_name} "
                f"{cur_layer_name} did not work: New weights do"
                f"not match pretrained weights"
            )
            num_layers_loaded += 1
            loaded_layer_names.append(cur_layer_name)
            if cfg.FREEZE_PRETRAINED:
                cur_layer.trainable = False
            else:
                cur_layer.trainable = True

    # Check that the right number of layers were loaded
    assert num_layers_loaded > 0, "Loading did not work: No weights were loaded!"

    if debug:
        pt_load_layers, cur_load_layers = get_layers_to_load(model, cfg)
        for cur_layer_tuple, pt_layer_tuple in zip(cur_load_layers, pt_load_layers):
            cur_layer_name = cur_layer_tuple[0]
            pt_layer = pt_layer_tuple[1]
            cur_layer = cur_layer_tuple[1]
            loaded_weight = pt_layer.get_weights()
            if loaded_weight:
                cur_weight = cur_layer.get_weights()
                assert np.all(loaded_weight[0] == cur_weight[0]), (
                    f"Loading weights for layer {cur_layer_name} in "
                    f"current model did not work: Model is not "
                    f"modified!"
                )
                if cfg.FREEZE_PRETRAINED:
                    assert not cur_layer.trainable, f"Layer '{cur_layer_name}' is not frozen!"
                else:
                    assert cur_layer.trainable, f"Layer '{cur_layer_name}' is frozen!"

    logger.info("All weights were successfully loaded!")
    logger.info(f"Loaded layer names in current model: {loaded_layer_names}")


class SelfSupervisedInfo(object):

    _LOADED_LAYER_NAMES = []

    @staticmethod
    def init_self_supervised(cfg: config.Config):
        # Check config format for self-supervised learning
        if cfg.LEARNING_TAG == "self-supervised":
            model = SelfSupervisedInfo.temp_build_model(cfg)
            if cfg.SS_LOADED_LAYERS:
                SelfSupervisedInfo._LOADED_LAYER_NAMES = cfg.SS_LOADED_LAYERS.copy()
            else:
                assert (
                    type(cfg.PRETRAINED_WEIGHTS_PATH) is dict
                ), "'PRETRAINED_WEIGHTS_PATH' must a dictionary"
                assert os.path.exists(cfg.PRETRAINED_CONFIG_PATH), (
                    f"'PRETRAINED_CONFIG_PATH' does not exist! "
                    f"Path: {cfg.PRETRAINED_CONFIG_PATH}"
                )

                # Get info of weights to be loaded
                pt_info = cfg.PRETRAINED_WEIGHTS_PATH
                weights_to_load = pt_info["weights"]

                # Ensure weights_to_load is of the right format for
                # self-supervised learning
                assert (
                    type(weights_to_load) is list
                ), "Key 'weights' of PRETRAINED_WEIGHTS_PATH must be a list"
                assert len(weights_to_load) == 1, (
                    "Must have only one weight dictionary defined in " "PRETRAINED_WEIGHTS_PATH"
                )
                weight_info = weights_to_load[0]
                assert (
                    "slice_indices" in weight_info
                ), "Key 'slice_indices' is required in weight dictionary"
                range_list = weight_info["slice_indices"]
                assert len(range_list) == 2, "The length of the list for key 'slice_indices' must 2"
                """
                assert range_list[1] == "until", \
                    "The last index of 'slice_indices' in weight dictionary " \
                    "must be the word 'until' for self-supervised learning"
                """

                # Get layers to be loaded
                _, total_cur_load_layers = get_layers_to_load(model, cfg)

                # Initialize _LOADED_LAYERS_NAMES
                for layer_tuple in total_cur_load_layers:
                    SelfSupervisedInfo._LOADED_LAYER_NAMES.append(layer_tuple[0])
            logger.info(
                f"SelfSupervisedInfo._LOADED_LAYER_NAMES: "
                f"{SelfSupervisedInfo._LOADED_LAYER_NAMES}"
            )

    @staticmethod
    def clear():
        SelfSupervisedInfo._LOADED_LAYER_NAMES.clear()

    @staticmethod
    def get_loaded_layer_names():
        return SelfSupervisedInfo._LOADED_LAYER_NAMES.copy()

    @staticmethod
    def temp_build_model(cfg):
        try:
            return build_model(cfg)
        except KeyError:
            return get_model(cfg)

    @staticmethod
    def set_to_inference(layer_name: str):
        if layer_name in SelfSupervisedInfo._LOADED_LAYER_NAMES:
            return {"training": False}
        else:
            return {}

    @staticmethod
    def create_self_supervised_model(model: Model, cfg: config.Config) -> Model:
        # Get instance of config used to train the pretrained model
        pt_info = cfg.PRETRAINED_WEIGHTS_PATH

        # Find all layers of current model
        cur_layer_list = get_all_layers(model)

        # Get weight info
        weights_to_load = pt_info["weights"]

        # Ensure weights_to_load is of the right format for
        # self-supervised learning
        assert (
            type(weights_to_load) is list
        ), "Key 'weights' of PRETRAINED_WEIGHTS_PATH must be a list"
        assert len(weights_to_load) == 1, (
            "Must have only one weight dictionary defined in " "PRETRAINED_WEIGHTS_PATH"
        )
        weight_info = weights_to_load[0]
        assert (
            "slice_indices" in weight_info
        ), "Key 'slice_indices' is required in weight dictionary"
        range_list = weight_info["slice_indices"]
        assert len(range_list) == 2, "The length of the list for key 'slice_indices' must 2"
        assert range_list[1] == "until", (
            "The last index of 'slice_indices' in weight dictionary "
            "must be the word 'until' for self-supervised learning"
        )

        cur_layer_name = ""
        found_name = False
        for layer_idx in range(len(cur_layer_list)):
            cur_layer_tuple = cur_layer_list[layer_idx]
            cur_layer_name = cur_layer_tuple[0]
            if word_contains_words(
                cur_layer_name, weight_info["include_words"], weight_info["exclude_words"]
            ):
                found_name = True
                break
        assert found_name, (
            "No layers found in current model that match description in " "weight dictionary"
        )

        # Split model
        split_layer = model.get_layer(name=cur_layer_name)
        if type(model.input) is list:
            pt_model_inputs = [Input(sub_input.shape) for sub_input in model.input]
        else:
            pt_model_inputs = Input(model.input.shape)
        pretrain_model = Model(
            inputs=pt_model_inputs, outputs=split_layer.input, name="pretrain_model"
        )
        if type(split_layer.input) is list:
            model_2_inputs = [Input(sub_input.shape) for sub_input in split_layer.input]
        else:
            model_2_inputs = Input(split_layer.input.shape)
        remainder_model = Model(inputs=model_2_inputs, outputs=model.output, name="remainder_model")
        # Define new model
        pretrain_model.trainable = not cfg.FREEZE_PRETRAINED
        x = pretrain_model(model.input, training=False)
        x = remainder_model(x)
        new_model = Model(inputs=model.input, outputs=x)

        return new_model

    @staticmethod
    def check_self_supervised_loading(pretrained_portion: Model, cfg: config.Config):
        """Checks that the weights of all the layers of the pretrained
        portion of the model will be loaded into the current model.
        """
        last_pt_layer_name = pretrained_portion.layers[-1].name

        # Get info of weights to be loaded
        pt_info = cfg.PRETRAINED_WEIGHTS_PATH
        weights_to_load = pt_info["weights"]
        weight_info = weights_to_load[0]

        # Find all layers of pretrained portion of the model
        pt_layer_list = get_all_layers(pretrained_portion)

        # This will work because the pretrained portion is part
        # of the current model
        prev_layer_name = ""
        for layer_idx in range(len(pt_layer_list)):
            cur_layer_tuple = pt_layer_list[layer_idx]
            cur_layer_name = cur_layer_tuple[0]
            if word_contains_words(
                cur_layer_name, weight_info["include_words"], weight_info["exclude_words"]
            ):
                break
            prev_layer_name = cur_layer_name

        assert prev_layer_name == last_pt_layer_name, (
            f"Last layer loaded (name = {prev_layer_name}) does match last layer "
            f"of pretrained portion of current model "
            f"(name = {last_pt_layer_name})."
        )



"""Implementation of Fully-Convolutional Densenet.

Built from paper: https://arxiv.org/pdf/1611.09326.pdf
"""
from typing import Dict, Sequence, Union

import tensorflow as tf
from keras import backend as K
from keras.layers import Activation
from keras.layers import BatchNormalization as BN
from keras.layers import (
    Concatenate,
    Conv2D,
    Conv2DTranspose,
    Conv3D,
    Conv3DTranspose,
    Dropout,
    Input,
    MaxPooling2D,
    MaxPooling3D,
)
from keras.utils.layer_utils import get_source_inputs

from medsegpy.config import FCDenseNetConfig

from ..model import Model
from ..model_utils import add_sem_seg_activation
from .build import META_ARCH_REGISTRY, ModelBuilder


def build_fc_dense_block(
    x: tf.Tensor,
    filters: Union[int, Sequence[int]],
    kernel_size: Union[int, Sequence[int]] = 3,
    num_layers: int = 2,
    activation: str = "relu",
    kernel_initializer: Union[str, Dict] = "he_normal",
    dropout: float = 0.0,
    use_bn: bool = True,
    concat_input: bool = False,
    conv_type=Conv2D,
):
    """Builds fully convolutional dense block.

    The sequence is shown below.
    The blocks are repeated `` number of times.
    Layers in braces (`{}`) are optional.`(C)` refers to
    concatenation:

           -------------------------------------------  ---------
           |                                         |  |       |
           |                                         v  |       v
        x  -> {BN} -> {Act} -> Conv -> {Dropout} -> (C) -> ... (C) ... (C) ->
           |                                     |              |       ^
           |                                     |              |       |
           |                                     |              ---------
           |                                     |                      |
           |                                     |                      |
           |                                     ------------------------
           |                                                            |
           |                                                            |
           --------------------------------------------------------------

    Args:
        x (tf.Tensor): Input tensor.
        filters (int or Sequence[int]): Number of filters to use
            for each conv layer. If a sequence, will override number of conv
            layers specified by `num_layers`.
        kernel_size: Kernel size accepted by Keras convolution layers.
        num_layers (int, optional): Number of layers/subblocks to make.
        activation (str, optional): Activation type. If `""`, no activation is
            used.
        kernel_initializer: Kernel initializer accepted by
            `keras.layers.Conv(...)`.
        dropout (float, optional): Dropout rate.
        use_bn (bool, optional): If `True`, use batch normalization.
        concat_input (bool, optional): Concatenate input into dense block
            with final layer.
    """
    channel_axis = 1 if K.image_data_format() == "channels_first" else -1
    num_layers = len(filters) if isinstance(filters, Sequence) else num_layers
    assert isinstance(num_layers, int), "{}".format(type(num_layers))
    if isinstance(filters, int):
        filters = [filters] * num_layers

    stack = x
    layer_outputs = [x] if concat_input else []

    for i in range(num_layers):
        if use_bn:
            x = BN(axis=channel_axis, momentum=0.95, epsilon=0.001)(x)
        if activation:
            x = Activation(activation)(x)
        x = conv_type(
            filters[i], kernel_size, padding="same", kernel_initializer=kernel_initializer
        )(x)
        if dropout != 0:
            x = Dropout(rate=dropout)(x)

        layer_outputs.append(x)
        x = stack = Concatenate(axis=channel_axis)([stack, x])

    if len(layer_outputs) > 1:
        x = Concatenate(axis=channel_axis)(layer_outputs)
    else:
        x = layer_outputs[0]
    return x


@META_ARCH_REGISTRY.register()
class FCDenseNet(ModelBuilder):
    def __init__(self, cfg):
        super().__init__(cfg)
        if len(cfg.IMG_SIZE) == 3:
            self._pooler_type = MaxPooling2D
            self._conv_type = Conv2D
            self._conv_transpose_type = Conv2DTranspose
            self._dim = 2
        elif len(cfg.IMG_SIZE) == 4:
            self._pooler_type = MaxPooling3D
            self._conv_type = Conv3D
            self._conv_transpose_type = Conv3DTranspose
            self._dim = 3
        else:
            raise ValueError("Image size {} not supported. " "Must be 2D (HxWxC) or 3D (HxWxDxC)")

        self.kernel_size = (3, 3)

    def _get_enc_dec_counts(
        self, counts: Union[Sequence[int], Sequence[Sequence[int]]], depth: int, cfg_field: str
    ):
        """Returns the count for the encoder and decoder.

        Encoder count includes the bottleneck.
        """
        if len(counts) == 1:
            num_enc_counts = [counts[0]] * depth
            num_dec_counts = [counts[0]] * (depth - 1)
        elif len(counts) == depth:
            num_enc_counts = counts
            num_dec_counts = counts[-2::-1]
        elif len(counts) == 2 * depth - 1:
            num_enc_counts = counts[:depth]
            num_dec_counts = counts[depth:]
        else:
            raise ValueError(
                "len(cfg.{}) should be 1, cfg.DEPTH, or 2*cfg.DEPTH-1. "
                "cfg.DEPTH={}, len(cfg.NUM_FILTERS)={}".format(cfg_field, depth, len(counts))
            )
        return num_enc_counts, num_dec_counts

    def build_model(self, input_tensor=None):
        cfg: FCDenseNetConfig = self._cfg
        seed = cfg.SEED
        depth = cfg.DEPTH
        kernel_size = self.kernel_size
        num_filters = cfg.NUM_FILTERS
        num_layers = cfg.NUM_LAYERS
        dropout = cfg.DROPOUT
        enc_num_filters, dec_num_filters = self._get_enc_dec_counts(
            num_filters, depth, "NUM_FILTERS"
        )
        enc_num_layers, dec_num_layers = self._get_enc_dec_counts(num_layers, depth, "NUM_LAYERS")
        kernel_initializer = {"class_name": cfg.KERNEL_INITIALIZER, "config": {"seed": seed}}
        output_mode = cfg.LOSS[1]
        assert output_mode in ["sigmoid", "softmax"]
        num_classes = cfg.get_num_classes()

        conv_type = self._conv_type
        pool_type = self._pooler_type
        ct_type = self._conv_transpose_type
        channel_axis = 1 if K.image_data_format() == "channels_first" else -1

        # Build inputs.
        if not input_tensor:
            inputs = self._build_input(cfg.IMG_SIZE)
        else:
            inputs = input_tensor

        x = inputs
        x = conv_type(
            cfg.NUM_FILTERS_HEAD_CONV,
            kernel_size,
            padding="same",
            kernel_initializer=kernel_initializer,
        )(x)

        # Encoder.
        skip_connections = []
        running_n_filters = [cfg.NUM_FILTERS_HEAD_CONV]
        running_pool_sizes = []
        for i in range(depth - 1):
            n_filters = enc_num_filters[i]
            n_layers = enc_num_layers[i]
            x = build_fc_dense_block(
                x,
                filters=n_filters,
                kernel_size=kernel_size,
                num_layers=n_layers,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=dropout,
                use_bn=True,
                concat_input=True,
                conv_type=conv_type,
            )
            skip_connections.append(x)
            if isinstance(n_filters, Sequence):
                running_n_filters.append(sum(n_filters))
            else:
                running_n_filters.append(n_filters * n_layers)

            # Transition down.
            x = build_fc_dense_block(
                x,
                filters=sum(running_n_filters),
                kernel_size=1,
                num_layers=1,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=dropout,
                use_bn=True,
                concat_input=False,
                conv_type=conv_type,
            )
            pool_size = self._get_pool_size(x)
            running_pool_sizes.append(pool_size)
            x = pool_type(pool_size=pool_size)(x)

        # Bottleneck.
        n_filters = enc_num_filters[depth - 1]
        n_layers = enc_num_layers[depth - 1]
        x = build_fc_dense_block(
            x,
            filters=n_filters,
            kernel_size=kernel_size,
            num_layers=n_layers,
            activation="relu",
            kernel_initializer=kernel_initializer,
            dropout=dropout,
            use_bn=True,
            concat_input=False,
            conv_type=self._conv_type,
        )
        if isinstance(n_filters, Sequence):
            running_n_filters.append(sum(n_filters))
        else:
            running_n_filters.append(n_filters * n_layers)

        # Decoder.
        running_n_filters = running_n_filters[:0:-1]  # remove head
        skip_connections = skip_connections[::-1]
        running_pool_sizes = running_pool_sizes[::-1]
        for i in range(depth - 1):
            # Transition Up.
            x = ct_type(
                filters=running_n_filters[i],
                kernel_size=kernel_size,
                padding="same",
                strides=running_pool_sizes[i],
                kernel_initializer=kernel_initializer,
            )(x)
            # TODO: add padding layer.

            x = Concatenate(axis=channel_axis)([skip_connections[i], x])

            n_filters = dec_num_filters[i]
            n_layers = dec_num_filters[i]
            x = build_fc_dense_block(
                x,
                filters=n_filters,
                kernel_size=kernel_size,
                num_layers=n_layers,
                activation="relu",
                kernel_initializer=kernel_initializer,
                dropout=dropout,
                use_bn=True,
                concat_input=False,
                conv_type=self._conv_type,
            )

        # 1x1 convolution to get pixel-wise semantic segmentation.
        x = add_sem_seg_activation(
            x,
            num_classes,
            conv_type=self._conv_type,
            activation=output_mode,
            kernel_initializer="glorot_uniform",
            seed=seed,
        )

        if input_tensor is not None:
            inputs = get_source_inputs(input_tensor)

        model = Model(inputs=inputs, outputs=[x])

        return model

    def _get_pool_size(self, x: tf.Tensor):
        return [2 if d % 2 == 0 or d % 3 != 0 else 3 for d in x.shape.as_list()[1:-1]]

    def _build_input(self, input_size):
        if len(input_size) == 2:
            input_size = input_size + (1,)
        elif len(input_size) > 4:
            raise ValueError("Input size must be 2D (HxW or HxWxC) or 3D (HxWxDxC)")

        return Input(input_size)



"""Overload Keras deserialization for default with custom layers and model names."""
from keras.layers import deserialize as _deserialize

from medsegpy.utils import env

from .pooling import MaxPoolingWithArgmax2D, MaxUnpooling2D
from .upsampling import BilinearUpsampling

if env.is_tf2():
    from tensorflow.python.keras.engine.functional import Functional
else:
    Functional = None


def deserialize(config, custom_objects=None):
    """Overload keras.layers.deserialize with MedSegPy builtin layers.

    Args:
      config: dict of the form {'class_name': str, 'config': dict}
      custom_objects: dict mapping class names (or function names)
          of custom (non-Keras) objects to class/functions

    Returns:
      Layer instance (may be Model, Sequential, Network, Layer...)
    """
    # Prevent circular dependencies.
    from ..model import Model  # noqa

    all_custom_objects = {
        "Model": Model,
        "BilinearUpsampling": BilinearUpsampling,
        "MaxPoolingWithArgmax2D": MaxPoolingWithArgmax2D,
        "MaxUnpooling2D": MaxUnpooling2D,
    }
    all_custom_objects.update(custom_objects)
    model = _deserialize(config, custom_objects=all_custom_objects)
    # if env.is_tf2() and isinstance(model, Functional):
    #     model = Model(model)
    return model



import warnings
from abc import ABC, abstractmethod

from fvcore.common.registry import Registry

from medsegpy.config import Config

from ..model import Model

META_ARCH_REGISTRY = Registry("META_ARCH")
META_ARCH_REGISTRY.__doc__ = """
Registry for meta-architectures, i.e. the whole model.

The registered object will be called with `obj(cfg)`. The resulting object
should be duck typed with `build_model(input_tensor)`.
"""

_MODEL_MAP = {
    ("unet_2d", "unet_2_5d"): "UNet2D",
    ("unet_3d",): "UNet3D",
    ("deeplabv3_2d", "deeplabv3_2_5d", "deeplabv3+"): "DeeplabV3Plus",
}

LEGACY_MODEL_NAMES = {x: v for k, v in _MODEL_MAP.items() for x in k}


def build_model(cfg, input_tensor=None) -> Model:
    """
    Build the whole model architecture, defined by ``cfg.MODEL_NAME``.
    Note that it does not load any weights from ``cfg``.
    """
    name = cfg.MODEL_NAME
    try:
        META_ARCH_REGISTRY.get(name)
    except KeyError:
        # Legacy code used different tags for building models.
        prev_name = name
        if name in LEGACY_MODEL_NAMES:
            name = LEGACY_MODEL_NAMES[name]
            if prev_name != name:
                warnings.warn("MODEL_NAME {} is deprecated. Use {} instead".format(prev_name, name))

    builder = META_ARCH_REGISTRY.get(name)(cfg)
    model = builder.build_model(input_tensor)
    assert isinstance(model, Model), (
        "ModelBuilder.build_model should output model of type " "medsegpy.modeling.Model"
    )
    return model


class ModelBuilder(ABC):
    def __init__(self, cfg: Config):
        self._cfg = cfg

    @abstractmethod
    def build_model(self, input_tensor=None) -> Model:
        """Build model."""
        pass



import warnings

import numpy as np
from keras.initializers import glorot_uniform, he_normal
from keras.layers import Activation
from keras.layers import BatchNormalization as BN
from keras.layers import Conv3D, Deconvolution3D, Dropout, Input, MaxPooling3D, concatenate
from keras.models import Model
from keras.utils import plot_model

DEFAULT_INPUT_SIZE = (288, 288, 64, 1)


def unet_3d_model(
    input_size=DEFAULT_INPUT_SIZE,
    input_tensor=None,
    num_classes=1,
    activation="sigmoid",
    output_mode=None,
    num_filters=None,
    depth=6,
    filter_size=(3, 3, 3),
    in_plane_pool_size=(2, 2),
    seed=None,
):
    warnings.warn("unet_3d_model is deprecated. Use UNet3D.build_model()", DeprecationWarning)
    # input size is a tuple of the size of the image
    # assuming channel last
    # input_size = (dim1, dim2, dim3, ch)
    # unet begins
    if input_tensor is None and (type(input_size) is not tuple or len(input_size) != 4):
        raise ValueError("input_size must be a tuple of size (height, width, slices, 1)")

    if num_filters is None:
        nfeatures = [2**feat * 32 for feat in np.arange(depth)]
    else:
        nfeatures = num_filters
        assert len(nfeatures) == depth

    xres, yres, slices, chans = input_size

    conv_ptr = []
    dim3 = []

    # Calculate what depth pooling should stop in dim3
    for _ in range(depth):
        slices /= 2
        dim3.append(int(np.floor(slices)))

    # input layer
    inputs = Input(input_size)

    # step down convolutional layers
    pool = inputs
    for depth_cnt in range(depth):

        conv = Conv3D(
            nfeatures[depth_cnt],
            filter_size,
            padding="same",
            kernel_initializer=he_normal(seed=seed),
        )(pool)
        conv = Activation("relu")(conv)
        conv = Conv3D(
            nfeatures[depth_cnt],
            filter_size,
            padding="same",
            kernel_initializer=he_normal(seed=seed),
        )(conv)
        conv = Activation("relu")(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.0)(conv)

        conv_ptr.append(conv)

        # Pool in slice dimension only if dim has length => 2
        if dim3[depth_cnt] > 0:
            pool = MaxPooling3D(pool_size=in_plane_pool_size + (2,))(conv)
        else:
            pool = MaxPooling3D(pool_size=in_plane_pool_size + (1,))(conv)

    # step up convolutional layers
    for depth_cnt in range(depth - 2, -1, -1):

        deconv_shape = conv_ptr[depth_cnt].shape.as_list()
        deconv_shape[0] = None

        # Deconv in slice dimension only if slice dim has size > 2
        if dim3[depth_cnt] > 0:
            up = concatenate(
                [
                    Deconvolution3D(
                        nfeatures[depth_cnt],
                        filter_size,
                        padding="same",
                        strides=in_plane_pool_size + (2,),
                        # output_shape=deconv_shape,
                        kernel_initializer=he_normal(seed=seed),
                    )(conv),
                    conv_ptr[depth_cnt],
                ],
                axis=4,
            )

        else:
            up = concatenate(
                [
                    Deconvolution3D(
                        nfeatures[depth_cnt],
                        filter_size,
                        padding="same",
                        strides=in_plane_pool_size + (1,),
                        # output_shape=deconv_shape,
                        kernel_initializer=he_normal(seed=seed),
                    )(conv),
                    conv_ptr[depth_cnt],
                ],
                axis=4,
            )

        conv = Conv3D(
            nfeatures[depth_cnt],
            filter_size,
            padding="same",
            kernel_initializer=he_normal(seed=seed),
        )(up)
        conv = Activation("relu")(conv)
        conv = Conv3D(
            nfeatures[depth_cnt],
            filter_size,
            padding="same",
            kernel_initializer=he_normal(seed=seed),
        )(conv)
        conv = Activation("relu")(conv)

        conv = BN(axis=-1, momentum=0.95, epsilon=0.001)(conv)
        conv = Dropout(rate=0.00)(conv)

    # combine features
    recon = Conv3D(
        num_classes, (1, 1, 1), padding="valid", kernel_initializer=glorot_uniform(seed=seed)
    )(conv)
    recon = Activation(activation)(recon)

    model = Model(inputs=[inputs], outputs=[recon])

    return model


if __name__ == "__main__":
    m = unet_3d_model(input_size=(288, 288, 16, 1))
    plot_model(m, "./imgs/unet_3d.png", show_shapes=True)



from .utils.env import setup_environment

setup_environment()


# This line will be programatically read/write by setup.py.
# Leave them at the bottom of this file and don't touch them.
__version__ = "0.0.1"



"""
THIS FILE IS DEPRECATED. DO NOT ADD ANY NEW MODELS HERE.
"""
import logging
import warnings

from keras.initializers import glorot_uniform
from keras.layers import Conv2D, Input

from medsegpy.config import (
    AnisotropicUNetConfig,
    DeeplabV3_2_5DConfig,
    DeeplabV3Config,
    RefineNetConfig,
    ResidualUNet,
    SegnetConfig,
    UNet2_5DConfig,
    UNet3DConfig,
    UNetConfig,
)

from .deeplab_2d.deeplab_model import DeeplabModel
from .model import Model
from .refinenet.refinenet_model import refinenet_model
from .segnet_2d.segnet import Segnet_v2
from .segnet_2d.segnet_bottleneck import SegNetBottleneck
from .unet_2d.anisotropic_unet_model import anisotropic_unet_2d
from .unet_2d.residual_unet_model import residual_unet_2d
from .unet_2d.unet_model import unet_2d_model, unet_2d_model_v2
from .unet_3d_model import unet_3d_model

warnings.warn(
    "Module medsegpy.modeling.build is deprecated. Use medsegpy.modeling.meta_arch.build.",
    DeprecationWarning,
)


logger = logging.getLogger(__name__)


def get_model(config):
    """
    Get model corresponding to config
    :param config: Config object with specific architecture configurations
    :return: a Keras model
    """
    if type(config) is DeeplabV3Config:
        model = deeplabv3_2d(config)
    elif type(config) is SegnetConfig:
        model = segnet_2d(config)
    elif type(config) is UNetConfig:
        model = unet_2d(config)
    elif type(config) is UNet2_5DConfig:
        model = unet_2_5d(config)
    elif type(config) is DeeplabV3_2_5DConfig:
        model = deeplabv3_2_5d(config)
    elif type(config) is ResidualUNet:
        model = residual_unet(config)
    elif type(config) is AnisotropicUNetConfig:
        model = anisotropic_unet(config)
    elif type(config) is RefineNetConfig:
        model = basic_refinenet(config)
    elif type(config) is UNet3DConfig:
        model = unet_3d(config)
    else:
        raise ValueError("This config type has not been implemented")

    # if weighted cross entropy, use softmax
    return model


def basic_refinenet(config):
    input_shape = config.IMG_SIZE
    activation = config.LOSS[1]
    num_classes = config.get_num_classes()
    model = refinenet_model(input_shape=input_shape)

    # Add activation
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    return model


def anisotropic_unet(config):
    input_shape = config.IMG_SIZE
    activation = config.LOSS[1]
    num_classes = config.get_num_classes()

    DEPTH = config.DEPTH
    NUM_FILTERS = config.NUM_FILTERS
    model = anisotropic_unet_2d(
        input_size=input_shape, depth=DEPTH, num_filters=NUM_FILTERS, kernel_size=config.KERNEL_SIZE
    )

    # Add activation
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    return model


def residual_unet(config):
    """
    Returns ResidualUnet model
    :param config:
    :return:
    """
    input_shape = config.IMG_SIZE
    activation = config.LOSS[1]
    num_classes = config.get_num_classes()

    DEPTH = config.DEPTH
    NUM_FILTERS = config.NUM_FILTERS
    model = residual_unet_2d(
        input_size=input_shape,
        depth=DEPTH,
        num_filters=NUM_FILTERS,
        layer_order=config.LAYER_ORDER,
        dropout_rate=config.DROPOUT_RATE,
        use_squeeze_excitation=config.USE_SE_BLOCK,
        squeeze_excitation_ratio=config.SE_RATIO,
    )

    # Add activation
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    return model


def unet_3d(config):
    """
    Returns Unet3D model
    :param config: a UNetConfig object
    :return: a Keras model

    :raises ValueError: if config not of type UNetConfig
    """
    input_shape = config.IMG_SIZE
    activation = config.LOSS[1]
    num_classes = config.get_num_classes()

    DEPTH = config.DEPTH
    NUM_FILTERS = config.NUM_FILTERS
    model = unet_3d_model(
        input_size=input_shape,
        depth=DEPTH,
        num_filters=NUM_FILTERS,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )

    return model


def unet_2d(config):
    """
    Returns Unet2D model
    :param config: a UNetConfig object
    :return: a Keras model

    :raises ValueError: if config not of type UNetConfig
    """
    warnings.warn("unet_2d is deprecated. Use `meta_arch.build_model(...)`.", DeprecationWarning)
    input_shape = config.IMG_SIZE
    activation = config.LOSS[1]
    num_classes = config.get_num_classes()

    # Legacy: some weights were trained on different structure
    # (conv and activation were combined) making loading weights difficult.
    # We check to see if we are testing and if we are in a case where we need
    # to account for this issue.
    if config.STATE == "testing" and config.VERSION <= 2:
        model = unet_2d_model(input_size=input_shape, output_mode="sigmoid")
    else:
        DEPTH = config.DEPTH
        NUM_FILTERS = config.NUM_FILTERS
        model = unet_2d_model_v2(input_size=input_shape, depth=DEPTH, num_filters=NUM_FILTERS)

        # Add activation
        x = __add_activation_layer(
            output=model.layers[-1].output,
            num_classes=num_classes,
            activation=activation,
            seed=config.SEED,
        )
        model = Model(inputs=model.input, outputs=x)

    return model


def deeplabv3_2d(config):
    """
    Returns DeeplabV3+ model
    :param config: a DeeplabV3Config object
    :return: a Keras model

    :raises ValueError: if config not of type DeeplabV3Config
    """
    warnings.warn(
        "deeplabv3_2d is deprecated. Use `meta_arch.build_model(...)`.", DeprecationWarning
    )

    if type(config) is not DeeplabV3Config:
        raise ValueError("config must be an instance of DeeplabV3Config")

    input_shape = config.IMG_SIZE
    OS = config.OS
    dil_rate_input = config.DIL_RATES
    activation = config.LOSS[1]
    dropout_rate = config.DROPOUT_RATE
    num_classes = config.get_num_classes()
    m = DeeplabModel(kernel_initializer=config.KERNEL_INITIALIZER, seed=config.SEED)
    model = m.Deeplabv3(
        weights=None,
        input_shape=input_shape,
        classes=num_classes,
        backbone="xception",
        OS=OS,
        dil_rate_input=dil_rate_input,
        dropout_rate=dropout_rate,
    )

    # Add sigmoid activation layer -
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    return model


def segnet_2d(config: SegnetConfig):
    """
    Returns SegnetConfig model
    :param config: a SegnetConfig object
    :return: a Keras model

    :raises ValueError: if config not of type SegnetConfig
    """
    if type(config) is not SegnetConfig:
        raise ValueError("config must be an instance of SegnetConfig")
    num_classes = config.get_num_classes()
    input_shape = config.IMG_SIZE
    output_mode = config.LOSS[1]
    if config.USE_BOTTLENECK:
        model = SegNetBottleneck(
            input_shape=input_shape,
            n_labels=num_classes,
            depth=config.DEPTH,
            num_conv_layers=config.NUM_CONV_LAYERS,
            num_filters=config.NUM_FILTERS,
            single_bn=config.SINGLE_BN,
            conv_act_bn=config.CONV_ACT_BN,
            output_mode=output_mode,
            seed=config.SEED,
        )
        model = model.build_model()
    else:
        model = Segnet_v2(
            input_shape=input_shape,
            n_labels=num_classes,
            depth=config.DEPTH,
            num_conv_layers=config.NUM_CONV_LAYERS,
            num_filters=config.NUM_FILTERS,
            single_bn=config.SINGLE_BN,
            conv_act_bn=config.CONV_ACT_BN,
            output_mode=output_mode,
            seed=config.SEED,
        )

    return model


def unet_2_5d(config):
    """
    Returns unet model corresponding to 3-channel multi-contrast inputs
    :param config: a UNetMultiContrastConfig object
    :return: a Keras model

    :raises ValueError: if config not of type UNetMultiContrastConfig
    """
    warnings.warn("unet_2_5d is deprecated. Use `meta_arch.build_model(...)`.", DeprecationWarning)
    if type(config) is not UNet2_5DConfig:
        raise ValueError("config must be instance of UNet2_5DConfig")

    activation = config.LOSS[1]
    num_classes = config.get_num_classes()
    input_shape = config.IMG_SIZE

    logger.info("Initializing 2.5d unet: input size - " + str(input_shape))

    x = Input(input_shape)

    model = unet_2d_model_v2(input_tensor=x)

    # Add activation
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    # only load weights for layers that share the same name
    if config.INIT_UNET_2D:
        model.load_weights(config.INIT_UNET_2D_WEIGHTS, by_name=True)

    return model


def deeplabv3_2_5d(config):
    """
    Returns unet model corresponding to 3-channel multi-contrast inputs
    :param config: a UNetMultiContrastConfig object
    :return: a Keras model

    :raises ValueError: if config not of type UNetMultiContrastConfig
    """
    warnings.warn(
        "deeplabv3_2_5d is deprecated. Use `meta_arch.build_model(...)`.", DeprecationWarning
    )
    if type(config) is not DeeplabV3_2_5DConfig:
        raise ValueError("config must be instance of DeeplabV3_2_5DConfig")
    logger.info("Initializing 2.5d deeplab: input size - " + str(config.IMG_SIZE))

    input_shape = config.IMG_SIZE
    OS = config.OS
    dil_rate_input = config.DIL_RATES
    activation = config.LOSS[1]
    dropout_rate = config.DROPOUT_RATE
    num_classes = config.get_num_classes()
    m = DeeplabModel(kernel_initializer=config.KERNEL_INITIALIZER, seed=config.SEED)
    model = m.Deeplabv3(
        weights=None,
        input_shape=input_shape,
        classes=num_classes,
        backbone="xception",
        OS=OS,
        dil_rate_input=dil_rate_input,
        dropout_rate=dropout_rate,
    )

    # Add sigmoid activation layer -
    x = __add_activation_layer(
        output=model.layers[-1].output,
        num_classes=num_classes,
        activation=activation,
        seed=config.SEED,
    )
    model = Model(inputs=model.input, outputs=x)

    return model


def __add_activation_layer(output, num_classes, activation="sigmoid", seed=None):
    """
    Return sigmoid activation layer
    :param: output: The output of the previous layer
    """

    # Initializing kernel weights to 1 and bias to 0.
    # i.e. without training, the output would be a sigmoid activation on each
    # pixel of the input
    return Conv2D(
        num_classes,
        (1, 1),
        activation=activation,
        kernel_initializer=glorot_uniform(seed=seed),
        name="output_activation",
    )(output)



#!/usr/bin/env python

import os
import sys
from os import path
from shutil import rmtree

from setuptools import Command, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py upload.
    Adapted from https://github.com/robustness-gym/meerkat.
    """

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(get_version()))
        os.system("git push --tags")

        sys.exit()


def get_version():
    init_py_path = path.join(path.abspath(path.dirname(__file__)), "medsegpy", "__init__.py")
    init_py = open(init_py_path, "r").readlines()
    version_line = [l.strip() for l in init_py if l.startswith("__version__")][0]
    version = version_line.split("=")[-1].strip().strip("'\"")

    # The following is used to build release packages.
    # Users should never use it.
    suffix = os.getenv("MEDSEGPY_VERSION_SUFFIX", "")
    version = version + suffix
    if os.getenv("BUILD_NIGHTLY", "0") == "1":
        from datetime import datetime

        date_str = datetime.today().strftime("%y%m%d")
        version = version + ".dev" + date_str

        new_init_py = [l for l in init_py if not l.startswith("__version__")]
        new_init_py.append('__version__ = "{}"\n'.format(version))
        with open(init_py_path, "w") as f:
            f.write("".join(new_init_py))
    return version


def get_required_packages():
    """Returns list of required packages based on tensorflow and keras versions."""
    default = [
        "Pillow",  # you can also use pillow-simd for better performance
        "matplotlib",
        "seaborn",
        "mock",
        "fvcore",
        "pydot",
        "pandas",
        "medpy",
        "numpy",
        "h5py",
        "natsort",
        "scipy",
        "scikit-image",
        "simpleitk",
        "configparser",
        "resnet",
        "more-itertools",
    ]
    return default


setup(
    name="medsegpy",
    version=get_version(),
    author="Arjun Desai",
    url="https://github.com/ad12/MedSegPy",
    description="MedSegPy is a framework for research on medical image segmentation.",
    packages=find_packages(exclude=("configs", "tests", "*.tests", "*.tests.*", "tests.*")),
    python_requires=">=3.6",
    install_requires=get_required_packages(),
    extras_require={
        "all": ["shapely", "psutil"],
        "dev": [
            "flake8==4.0.1",
            "isort",
            "black==22.8.0",
            "flake8-bugbear",
            "flake8-comprehensions",
            "sphinx-rtd-theme",
            "nbsphinx",
            "recommonmark",
        ],
    },
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



"""Normalization Layers

The following layers are useful for normalizing the inputs to
layers.

1) Group Normalization:
    - Copied from keras_contrib/layers/normalization/groupnormalization.py
    - Replaced the call to KC.moments with a call to the actual tensorflow function
    - Modified 'get_config()' to match structure of 'get_config()' in ./attention.py
"""
import tensorflow as tf
from keras import backend as K
from keras import constraints, initializers, regularizers
from keras.layers import InputSpec, Layer


class GroupNormalization(Layer):
    """Group normalization layer.

    Arguments
        groups: Integer, the number of groups for Group Normalization.
            Can be in the range [1, N] where N is the input dimension.
            The input dimension must be divisible by the number of groups.
        axis: Integer, the axis that should be normalized
            (typically the features axis).
            For instance, after a `Conv2D` layer with
            `data_format="channels_first"`,
            set `axis=1` in `BatchNormalization`.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
            When the next layer is linear (also e.g. `nn.relu`),
            this can be disabled since the scaling
            will be done by the next layer.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.

    Reference:
        Wu, Yuxin and He, Kaiming. "Group normalization". ECCV. 2018.
    """

    def __init__(
        self,
        groups=32,
        axis=-1,
        momentum=0.99,
        epsilon=1e-5,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs
    ):
        super(GroupNormalization, self).__init__(**kwargs)
        self.supports_masking = True
        self.groups = groups
        self.axis = axis
        self.momentum = momentum
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = initializers.get(beta_initializer)
        self.gamma_initializer = initializers.get(gamma_initializer)
        self.moving_mean_initializer = initializers.get(moving_mean_initializer)
        self.moving_variance_initializer = initializers.get(moving_variance_initializer)
        self.beta_regularizer = regularizers.get(beta_regularizer)
        self.gamma_regularizer = regularizers.get(gamma_regularizer)
        self.beta_constraint = constraints.get(beta_constraint)
        self.gamma_constraint = constraints.get(gamma_constraint)

    def build(self, input_shape):
        dim = input_shape[self.axis]

        if dim is None:
            raise ValueError(
                "Axis " + str(self.axis) + " of "
                "input tensor should have a defined dimension "
                "but the layer received an input with shape " + str(input_shape) + "."
            )

        if dim < self.groups:
            raise ValueError(
                "Number of groups (" + str(self.groups) + ") cannot be "
                "more than the number of channels (" + str(dim) + ")."
            )

        if dim % self.groups != 0:
            raise ValueError(
                "Number of groups (" + str(self.groups) + ") must be a "
                "multiple of the number of channels (" + str(dim) + ")."
            )

        self.input_spec = InputSpec(ndim=len(input_shape), axes={self.axis: dim})
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(
                shape=shape,
                name="gamma",
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
            )
        else:
            self.gamma = None
        if self.center:
            self.beta = self.add_weight(
                shape=shape,
                name="beta",
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
            )
        else:
            self.beta = None

        mean_var_shape = (self.groups,)

        self.moving_mean = self.add_weight(
            shape=mean_var_shape,
            name="moving_mean",
            initializer=self.moving_mean_initializer,
            trainable=False,
        )

        self.moving_variance = self.add_weight(
            shape=mean_var_shape,
            name="moving_variance",
            initializer=self.moving_variance_initializer,
            trainable=False,
        )

        self.built = True

    def call(self, inputs, training=None, **kwargs):
        input_shape = K.int_shape(inputs)
        tensor_input_shape = K.shape(inputs)

        # Prepare broadcasting shape.
        reduction_axes = list(range(len(input_shape)))
        del reduction_axes[self.axis]
        broadcast_shape = [1] * len(input_shape)
        broadcast_shape[self.axis] = input_shape[self.axis] // self.groups
        broadcast_shape.insert(1, self.groups)

        reshape_group_shape = K.shape(inputs)
        group_axes = [reshape_group_shape[i] for i in range(len(input_shape))]
        group_axes[self.axis] = input_shape[self.axis] // self.groups
        group_axes.insert(1, self.groups)

        # reshape inputs to new group shape
        group_shape = [group_axes[0], self.groups] + group_axes[2:]
        group_shape = K.stack(group_shape)
        inputs = K.reshape(inputs, group_shape)

        def normalize_inference(inputs):
            moving_broadcast_shape = [1] * len(input_shape)
            moving_broadcast_shape.insert(1, self.groups)

            # Broadcast moving mean and moving variance
            broadcast_moving_mean = K.reshape(self.moving_mean, moving_broadcast_shape)
            broadcast_moving_variance = K.reshape(self.moving_variance, moving_broadcast_shape)
            # Perform group normalization
            inputs = (inputs - broadcast_moving_mean) / (
                K.sqrt(broadcast_moving_variance + self.epsilon)
            )

            # prepare broadcast shape
            inputs = K.reshape(inputs, group_shape)
            outputs = inputs

            # In this case we must explicitly broadcast all parameters.
            if self.scale:
                broadcast_gamma = K.reshape(self.gamma, broadcast_shape)
                outputs = outputs * broadcast_gamma

            if self.center:
                broadcast_beta = K.reshape(self.beta, broadcast_shape)
                outputs = outputs + broadcast_beta

            # finally we reshape the output back to the input shape
            outputs = K.reshape(outputs, tensor_input_shape)

            return outputs

        # If the learning phase is *static* and set to inference:
        if training in {0, False}:
            return normalize_inference(inputs)

        group_reduction_axes = list(range(len(group_axes)))
        mean, variance = tf.nn.moments(inputs, axes=group_reduction_axes[2:], keep_dims=True)
        normed_inputs = (inputs - mean) / (K.sqrt(variance + self.epsilon))

        batch_mean = K.mean(mean, axis=0, keepdims=True)
        batch_variance = K.mean(variance, axis=0, keepdims=True)

        moving_shape = K.shape(self.moving_mean)
        batch_mean = K.reshape(batch_mean, moving_shape)
        batch_variance = K.reshape(batch_variance, moving_shape)

        if K.backend() != "cntk":
            sample_size = K.prod(
                [K.shape(normed_inputs)[axis] for axis in group_reduction_axes[2:]]
            )
            sample_size = K.cast(sample_size, dtype=K.dtype(normed_inputs))

            # sample variance - unbiased estimator of population variance
            batch_variance *= sample_size / (sample_size - (1.0 + self.epsilon))

        # Update moving mean and moving variance
        self.moving_mean = K.moving_average_update(self.moving_mean, batch_mean, self.momentum)
        self.moving_variance = K.moving_average_update(
            self.moving_variance, batch_variance, self.momentum
        )
        # prepare broadcast shape
        normed_inputs = K.reshape(normed_inputs, group_shape)
        outputs = normed_inputs

        # In this case we must explicitly broadcast all parameters.
        if self.scale:
            broadcast_gamma = K.reshape(self.gamma, broadcast_shape)
            outputs = outputs * broadcast_gamma

        if self.center:
            broadcast_beta = K.reshape(self.beta, broadcast_shape)
            outputs = outputs + broadcast_beta

        # finally we reshape the output back to the input shape
        outputs = K.reshape(outputs, tensor_input_shape)

        # Pick the normalized form corresponding to the training phase.
        return K.in_train_phase(outputs, lambda: normalize_inference(inputs), training=training)

    def get_config(self):
        base_config = super(GroupNormalization, self).get_config()
        base_config.update(
            {
                "groups": self.groups,
                "axis": self.axis,
                "epsilon": self.epsilon,
                "center": self.center,
                "scale": self.scale,
                "beta_initializer": initializers.serialize(self.beta_initializer),
                "gamma_initializer": initializers.serialize(self.gamma_initializer),
                "beta_regularizer": regularizers.serialize(self.beta_regularizer),
                "gamma_regularizer": regularizers.serialize(self.gamma_regularizer),
                "beta_constraint": constraints.serialize(self.beta_constraint),
                "gamma_constraint": constraints.serialize(self.gamma_constraint),
            }
        )
        return base_config

    def compute_output_shape(self, input_shape):
        return input_shape



import unittest

from medsegpy.config import DeeplabV3_2_5DConfig, DeeplabV3Config
from medsegpy.modeling.meta_arch import build
from medsegpy.utils import env

if not env.is_tf2():
    _TF2 = False
    from medsegpy.modeling.build import get_model
else:
    _TF2 = True
    get_model = None


class TestDeeplabV3Plus(unittest.TestCase):
    """Test building 2D/2.5D UNet using the builder."""

    def _compare_layers(self, m1, m2):
        for l1, l2 in zip(m1.layers, m2.layers):
            assert type(l1) == type(l2)  # noqa

            l1_cfg = l1.get_config()
            l2_cfg = l2.get_config()
            l2_cfg.pop("name"), l1_cfg.pop("name")

            assert l1_cfg == l2_cfg, "{}: {}\n{}: {}".format(
                l1.name, l1.get_config(), l2.name, l2.get_config()
            )

    def test_same_deeplabv3_2d(self):
        """
        2D DeeplabV3+ should be same between builder and function construct.
        """
        if _TF2:
            return
        cfg = DeeplabV3Config()
        cfg.CATEGORIES = [0]
        m1 = get_model(cfg)
        m2 = build.build_model(cfg)

        self._compare_layers(m1, m2)

    def test_same_deeplabv3_2_5d(self):
        """
        2.5D DeeplabV3+ should be same between builder and function construct.
        """
        if _TF2:
            return
        cfg1 = DeeplabV3_2_5DConfig()
        cfg1.IMG_SIZE = (288, 288, 3)
        cfg1.CATEGORIES = [0]
        cfg2 = DeeplabV3Config()
        cfg2.IMG_SIZE = (288, 288, 3)
        cfg2.CATEGORIES = [0]
        m1 = get_model(cfg1)
        m2 = build.build_model(cfg2)

        self._compare_layers(m1, m2)



from .build import build_loader
from .catalog import DatasetCatalog, MetadataCatalog
from .data_loader import (
    DATA_LOADER_REGISTRY,
    DataLoader,
    DefaultDataLoader,
    N5dDataLoader,
    PatchDataLoader,
)

# ensure the builtin datasets are registered
from .im_gens import (
    CTGenerator,
    Generator,
    GeneratorState,
    OAI3DBlockGenerator,
    OAI3DGenerator,
    OAI3DGeneratorFullVolume,
    OAIGenerator,
    get_generator,
)

from . import datasets  # isort:skip

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import unittest

import keras.backend as K
import numpy as np

from medsegpy.loss.utils import get_activation, get_shape, reduce_tensor, to_numpy

try:
    import tf.keras.activations as activations
except ImportError:
    import keras.activations as activations

X = np.asarray([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]])
WEIGHTS = np.asarray([[1, 2, 3, 4, 5]])


class TestGetShape(unittest.TestCase):
    def test_get_shape(self):
        x = K.constant(np.random.rand(2, 5, 9))
        assert get_shape(x) == (2, 5, 9)


class TestReduceTensor(unittest.TestCase):
    def test_basic_reduction(self):
        x = K.constant(X)

        val = to_numpy(reduce_tensor(x, reduction="mean"))
        assert np.allclose(val, np.mean(X))

        val = to_numpy(reduce_tensor(x, reduction="sum"))
        assert np.allclose(val, np.sum(X))

        val = to_numpy(reduce_tensor(x, reduction="none"))
        assert np.allclose(val, X)

    def test_weight(self):
        """Broadcasting weights should not affect weighted reduction calculations."""
        x = K.constant(X)  # Shape: (2, 5)

        w = np.asarray([1, 2, 3, 4, 5])  # Shape: (5,)
        weights = K.constant(w)

        val = to_numpy(reduce_tensor(x, weights=weights, reduction="mean"))
        assert np.allclose(val, np.mean(np.sum(w * X, axis=-1) / np.sum(WEIGHTS)))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="sum"))
        assert np.allclose(val, np.sum(w * X))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="none"))
        assert np.allclose(val, w * X)

        w = np.asarray([[1, 2, 3, 4, 5]])  # Shape: (1,5)
        weights = K.constant(w)

        val = to_numpy(reduce_tensor(x, weights=weights, reduction="mean"))
        assert np.allclose(val, np.mean(np.sum(w * X, axis=-1) / np.sum(WEIGHTS)))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="sum"))
        assert np.allclose(val, np.sum(w * X))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="none"))
        assert np.allclose(val, w * X)

        w = np.asarray([[1], [2]])  # Shape: (2,1)
        weights = K.constant(w)
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="mean"))
        assert np.allclose(val, np.sum(w * X) / (np.sum(w) * X.shape[1]))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="sum"))
        assert np.allclose(val, np.sum(w * X))
        val = to_numpy(reduce_tensor(x, weights=weights, reduction="none"))
        assert np.allclose(val, w * X)


class TestGetActivation(unittest.TestCase):
    def test_str(self):
        name, fn, args = get_activation("sigmoid")
        assert name == "sigmoid"
        assert fn == activations.sigmoid
        assert "axis" not in args

        name, fn, args = get_activation("softmax")
        assert name == "softmax"
        assert fn == activations.softmax
        assert "axis" in args

    def test_callable(self):
        name, fn, args = get_activation(activations.sigmoid)
        assert name == "sigmoid"
        assert fn == activations.sigmoid
        assert "axis" not in args

        name, fn, args = get_activation(activations.softmax)
        assert name == "softmax"
        assert fn == activations.softmax
        assert "axis" in args


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from medsegpy.loss.classification import DiceLoss
from medsegpy.loss.utils import to_numpy
from medsegpy.losses import avg_dice_loss, dice_loss, multi_class_dice_loss

try:
    import tf.keras.backend as K
except ImportError:
    import keras.backend as K

GT = np.asarray(
    [
        [[0, 1, 1, 0, 0], [1, 0, 0, 0, 0], [0, 0, 0, 1, 1]],
        [[0, 1, 1, 1, 0], [1, 1, 0, 0, 0], [1, 0, 0, 0, 0]],
        [[1, 1, 1, 0, 0], [0, 0, 0, 1, 0], [0, 0, 0, 0, 1]],
    ]
)
PRED = np.asarray(
    [
        [[0.1, 0.9, 0.9, 0.1, 0.1], [0.9, 0.1, 0.1, 0.1, 0.1], [0.1, 0.1, 0.1, 0.9, 0.9]],
        [[0.1, 0.9, 0.9, 0.1, 0.1], [0.9, 0.1, 0.1, 0.1, 0.1], [0.1, 0.1, 0.1, 0.9, 0.9]],
        [[1, 1, 1, 0, 0], [0, 0, 0, 1, 0], [0, 0, 0, 0, 1]],
    ]
)
EPS = 1e-8


class TestDiceLoss(unittest.TestCase):
    def test_flatten(self):
        gt = GT
        pred = PRED
        eps = EPS

        loss = DiceLoss(flatten=False, eps=eps, reduction=None)
        dim = 1
        expected = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (3, 5), val.shape
        assert np.allclose(val, expected)

        loss = DiceLoss(flatten="batch", eps=eps, reduction=None)
        dim = (0, 1)
        expected = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (5,), val.shape
        assert np.allclose(val, expected)

        loss = DiceLoss(flatten="channel", eps=eps, reduction=None)
        dim = (1, 2)
        expected = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (3,), val.shape
        assert np.allclose(val, expected)

        loss = DiceLoss(flatten=True, eps=eps, reduction=None)
        dim = (0, 1, 2)
        expected = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert np.isscalar(val)
        assert np.allclose(val, expected)

    def test_reduction(self):
        gt = GT
        pred = PRED
        eps = EPS

        dim = 1
        expected = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )

        loss = DiceLoss(reduction=None, eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (3, 5), val.shape
        assert np.allclose(val, expected)

        loss = DiceLoss(reduction="mean", eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert np.isscalar(val), val
        assert np.allclose(val, np.mean(expected))

        loss = DiceLoss(reduction="sum", eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert np.isscalar(val), val
        assert np.allclose(val, np.sum(expected))

    def test_weight(self):
        gt = GT
        pred = PRED
        eps = EPS

        weights = np.asarray([1.0, 2.0, 3.0, 4.0, 5.0])
        dim = 1
        dsc_loss = 1 - (
            2
            * (np.sum(gt * pred, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred, axis=dim) + eps)
        )
        dsc_loss[dsc_loss < eps] = 0
        expected = weights[np.newaxis, ...] * dsc_loss
        loss = DiceLoss(weights=weights, reduction=None, eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert np.allclose(val[2], expected[2])

    def test_activation(self):
        gt = GT
        pred = PRED
        eps = EPS

        dim = 1
        pred_sig = 1 / (1 + np.exp(-pred))
        expected = 1 - (
            2
            * (np.sum(gt * pred_sig, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred_sig, axis=dim) + eps)
        )
        loss = DiceLoss(activation="sigmoid", reduction=None, eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (3, 5), val.shape
        assert np.allclose(val, expected)

        dim = 1
        pred_soft = np.exp(pred) / np.sum(np.exp(pred), axis=-1, keepdims=True)
        expected = 1 - (
            2
            * (np.sum(gt * pred_soft, axis=dim) + eps)
            / (np.sum(gt, axis=dim) + np.sum(pred_soft, axis=dim) + eps)
        )
        loss = DiceLoss(activation="softmax", reduction=None, eps=eps)
        val = to_numpy(loss(K.constant(gt), K.constant(pred)))
        assert val.shape == (3, 5), val.shape
        assert np.allclose(val, expected)

    def test_reproducible(self):
        """
        Select loss functions in `medsegpy.losses` can be represented
        with new DiceLoss class.
        """
        gt = K.constant(GT)
        pred = K.constant(PRED)
        weights = np.asarray([1.0, 2.0, 3.0, 4.0, 5.0])

        # medsegpy.losses.dice_loss
        expected = to_numpy(dice_loss(gt, pred))
        loss = DiceLoss(reduction=None, flatten="channel")
        val = to_numpy(loss(gt, pred))
        assert np.allclose(val, expected)

        # medsegpy.losses.multi_class_dice_loss
        multi_class = multi_class_dice_loss(weights=weights, reduce="mean")
        expected = to_numpy(multi_class(gt, pred))
        loss = DiceLoss(weights=weights, reduction="mean", flatten="batch")
        val = to_numpy(loss(gt, pred))
        assert np.allclose(val, expected)

        multi_class = multi_class_dice_loss(reduce="class")
        expected = to_numpy(multi_class(gt, pred))
        loss = DiceLoss(reduction=None, flatten="batch")
        val = to_numpy(loss(gt, pred))
        assert np.allclose(val, expected)

        # medsegpy.losses.avg_dice_loss
        avg_dice = avg_dice_loss()
        expected = to_numpy(avg_dice(gt, pred))
        loss = DiceLoss(reduction="mean", flatten=False)
        val = to_numpy(loss(gt, pred))
        assert np.allclose(val, expected)

        avg_dice = avg_dice_loss(weights=weights)
        expected = to_numpy(avg_dice(gt, pred))
        loss = DiceLoss(weights=weights, reduction="mean", flatten=False)
        val = to_numpy(loss(gt, pred))
        assert np.allclose(val, expected)


if __name__ == "__main__":
    unittest.main()



import logging

import numpy as np
import scipy.special as sps
import tensorflow as tf
from keras import backend as K
from keras.callbacks import Callback
from keras.losses import binary_crossentropy

from medsegpy.loss.classification import DiceLoss
from medsegpy.utils import env

logger = logging.getLogger(__name__)

DICE_LOSS = ("dice", "sigmoid")
MULTI_CLASS_DICE_LOSS = ("multi_class_dice", "sigmoid")
AVG_DICE_LOSS = ("avg_dice", "sigmoid")
AVG_DICE_LOSS_SOFTMAX = ("avg_dice", "softmax")
AVG_DICE_NO_REDUCE = ("avg_dice_no_reduce", "sigmoid")
WEIGHTED_CROSS_ENTROPY_LOSS = ("weighted_cross_entropy", "softmax")
WEIGHTED_CROSS_ENTROPY_SIGMOID_LOSS = ("weighted_cross_entropy_sigmoid", "sigmoid")

BINARY_CROSS_ENTROPY_LOSS = ("binary_crossentropy", "softmax")

BINARY_CROSS_ENTROPY_SIG_LOSS = ("binary_crossentropy", "sigmoid")

FOCAL_LOSS = ("focal_loss", "sigmoid")
FOCAL_LOSS_GAMMA = 3.0

DICE_FOCAL_LOSS = ("dice_focal_loss", "sigmoid")
DICE_MEDIAN_LOSS = ("dice_median_loss", "sigmoid")

L2_LOSS = ("l2_loss", None)

CMD_LINE_SUPPORTED_LOSSES = [
    "DICE_LOSS",
    "MULTI_CLASS_DICE_LOSS",
    "AVG_DICE_LOSS",
    "AVG_DICE_LOSS_SOFTMAX",
    "AVG_DICE_NO_REDUCE",
    "WEIGHTED_CROSS_ENTROPY_LOSS",
    "WEIGHTED_CROSS_ENTROPY_SIGMOID_LOSS",
    "BINARY_CROSS_ENTROPY_LOSS",
    "BINARY_CROSS_ENTROPY_SIG_LOSS",
    "FOCAL_LOSS",
    "DICE_FOCAL_LOSS",
    "DICE_MEDIAN_LOSS",
    "L2_LOSS",
]


def build_loss(cfg):
    loss = cfg.LOSS
    num_classes = len(cfg.CATEGORIES)
    robust_loss_cls = cfg.ROBUST_LOSS_NAME
    robust_step_size = cfg.ROBUST_LOSS_STEP_SIZE

    if robust_loss_cls:
        reduction = "class"
    else:
        reduction = ""

    if isinstance(loss, str):
        try:
            loss = get_training_loss_from_str(loss)
        except (ValueError, AttributeError):
            pass
    loss = get_training_loss(
        loss,
        weights=cfg.CLASS_WEIGHTS,
        # Remove computation on the background class.
        remove_background=cfg.INCLUDE_BACKGROUND,
        reduce=reduction,
    )

    if not robust_loss_cls:
        return loss
    elif robust_loss_cls == "NaiveAdaRobLossComputer":
        return NaiveAdaRobLossComputer(
            criterion=loss, n_groups=num_classes, robust_step_size=robust_step_size
        )
    else:
        raise ValueError(f"{robust_loss_cls} not supported")


# TODO (arjundd): Add ability to exclude specific indices from loss function.
def get_training_loss_from_str(loss_str: str):
    loss_str = loss_str.upper()
    if loss_str == "DICE_LOSS":
        return DICE_LOSS
    elif loss_str == "MULTI_CLASS_DICE_LOSS":
        return MULTI_CLASS_DICE_LOSS
    elif loss_str == "AVG_DICE_LOSS" or loss_str == "AVG_DICE_LOSS_SOFTMAX":
        return AVG_DICE_LOSS
    elif loss_str == "AVG_DICE_NO_REDUCE":
        return AVG_DICE_NO_REDUCE
    elif loss_str == "WEIGHTED_CROSS_ENTROPY_LOSS":
        return WEIGHTED_CROSS_ENTROPY_LOSS
    elif loss_str == "WEIGHTED_CROSS_ENTROPY_SIGMOID_LOSS":
        return WEIGHTED_CROSS_ENTROPY_SIGMOID_LOSS
    elif loss_str == "BINARY_CROSS_ENTROPY_LOSS":
        return BINARY_CROSS_ENTROPY_LOSS
    elif loss_str == "BINARY_CROSS_ENTROPY_SIG_LOSS":
        return BINARY_CROSS_ENTROPY_SIG_LOSS
    elif loss_str == "FOCAL_LOSS":
        return FOCAL_LOSS
    elif loss_str == "DICE_FOCAL_LOSS":
        return DICE_FOCAL_LOSS
    elif loss_str == "DICE_MEDIAN_LOSS":
        return DICE_MEDIAN_LOSS
    elif loss_str == "L2_LOSS":
        return L2_LOSS
    else:
        raise ValueError("%s not supported" % loss_str)


def get_training_loss(loss, **kwargs):
    if loss == DICE_LOSS:
        return dice_loss
    elif loss == MULTI_CLASS_DICE_LOSS:
        return multi_class_dice_loss(**kwargs)
    elif loss == AVG_DICE_LOSS or loss == AVG_DICE_LOSS_SOFTMAX:
        return avg_dice_loss(**kwargs)
    elif loss == WEIGHTED_CROSS_ENTROPY_LOSS:
        return weighted_categorical_crossentropy(**kwargs)
    elif loss == BINARY_CROSS_ENTROPY_LOSS:
        return binary_crossentropy
    elif loss == BINARY_CROSS_ENTROPY_SIG_LOSS:
        return binary_crossentropy
    elif loss == FOCAL_LOSS:
        return focal_loss(**kwargs)
    elif loss == WEIGHTED_CROSS_ENTROPY_SIGMOID_LOSS:
        return weighted_categorical_crossentropy_sigmoid(**kwargs)
    elif loss == DICE_FOCAL_LOSS:
        return dice_focal_loss
    elif loss == DICE_MEDIAN_LOSS:
        return dice_median_loss
    elif loss == AVG_DICE_NO_REDUCE:
        kwargs.pop("reduce", None)
        kwargs["reduction"] = "none"
        return DiceLoss(**kwargs)
    elif loss == L2_LOSS:
        return l2_loss
    else:
        raise ValueError("Loss type not supported")


def _get_shape(x):
    """Returns shape of Keras tensor."""
    if hasattr(K, "int_shape"):
        return K.int_shape(x)
    else:
        return K.get_variable_shape(x)


def dice_focal_loss(y_true, y_pred):
    dsc = dice_loss(y_true, y_pred)
    fc = focal_loss(FOCAL_LOSS_GAMMA)(y_true, y_pred)

    return dsc + fc


# Dice function loss optimizer
def dice_loss(y_true, y_pred):
    """Computes class-agnostic dice."""
    szp = _get_shape(y_pred)

    img_len = np.product(szp[1:])

    if env.is_tf2():
        y_true = tf.dtypes.cast(y_true, y_pred.dtype)
    y_true = K.reshape(y_true, (-1, img_len))
    y_pred = K.reshape(y_pred, (-1, img_len))

    ovlp = K.sum(y_true * y_pred, axis=-1)

    mu = K.epsilon()
    dice = (2.0 * ovlp + mu) / (K.sum(y_true, axis=-1) + K.sum(y_pred, axis=-1) + mu)
    loss = 1 - dice

    return loss


def dice_median_loss(y_true, y_pred):
    """Get the median dice loss"""
    raise DeprecationWarning()
    lambda1 = 2
    mu = K.epsilon()

    szp = K.get_variable_shape(y_pred)
    img_len = szp[1] * szp[2] * szp[3]

    y_true = K.reshape(y_true, (-1, img_len))
    y_pred = K.reshape(y_pred, (-1, img_len))

    dsc = (2.0 * K.sum(y_true * y_pred, axis=-1) + mu) / (
        K.sum(y_true, axis=-1) + K.sum(y_pred, axis=-1) + mu
    )
    dsc_mean = K.mean(dsc)
    dsc_std = K.std(dsc)
    bool_mask = tf.logical_and(
        K.greater_equal(dsc - dsc_mean + lambda1 * dsc_std, 0),
        K.less_equal(dsc - dsc_mean - lambda1 * dsc_std, 0),
    )

    binarize_dsc = tf.boolean_mask(dsc, bool_mask)

    loss = 1 - K.mean(binarize_dsc)

    return loss


def multi_class_dice_loss(
    weights=None, remove_background: bool = False, reduce="mean", use_numpy=False, **kwargs
):
    """Dice loss for multiple classes in softmax layer.

    Each class is treated individually.
    """
    use_weights = False
    if weights is not None:
        weights = K.variable(weights)
        use_weights = True

    def d_loss(y_true, y_pred):
        szp = _get_shape(y_pred)

        if env.is_tf2():
            y_true = tf.dtypes.cast(y_true, y_pred.dtype)
        y_true = K.reshape(y_true, (-1, szp[-1]))
        y_pred = K.reshape(y_pred, (-1, szp[-1]))

        ovlp = K.sum(y_true * y_pred, axis=0)

        mu = K.epsilon()
        dice = (2.0 * ovlp + mu) / (K.sum(y_true, axis=0) + K.sum(y_pred, axis=0) + mu)
        loss = 1 - dice

        if reduce == "class":
            return loss

        if use_weights:
            loss = weights * loss
            loss = K.sum(loss) / K.sum(weights)
        else:
            loss = K.mean(loss)

        return loss

    def d_loss_np(y_true, y_pred):
        szp = y_pred.shape

        y_true = np.reshape(y_true, (-1, szp[-1]))
        y_pred = np.reshape(y_pred, (-1, szp[-1]))

        ovlp = np.sum(y_true * y_pred, axis=0)

        mu = K.epsilon()
        dice = (2.0 * ovlp + mu) / (np.sum(y_true, axis=0) + np.sum(y_pred, axis=0) + mu)
        loss = 1 - dice

        if reduce == "class":
            return loss

        if use_weights:
            loss = weights * loss
            loss = np.sum(loss) / np.sum(weights)
        else:
            loss = np.mean(loss)

        return loss

    if use_numpy:
        return d_loss_np
    else:
        return d_loss


def avg_dice_loss(weights=None, remove_background: bool = False, **kwargs):
    use_weights = False
    if weights is not None:
        weights = np.asarray(weights)[np.newaxis, ...]
        weights = K.variable(weights)
        use_weights = True

    def d_loss(y_true, y_pred):
        szp = _get_shape(y_pred)
        c_dim = szp[-1]  # class dimension
        img_size = np.prod(szp[1:-1])  # vectorized image size

        # Keep batch and class dimensions.
        y_true = K.reshape(y_true, (-1, img_size, c_dim))
        y_pred = K.reshape(y_pred, (-1, img_size, c_dim))
        if remove_background:
            y_true = y_true[..., 1:]
            y_pred = y_pred[..., 1:]

        ovlp = K.sum(y_true * y_pred, axis=1)

        mu = K.epsilon()
        dice = (2.0 * ovlp + mu) / (K.sum(y_true, axis=1) + K.sum(y_pred, axis=1) + mu)
        loss = 1 - dice

        if use_weights:
            loss = weights * loss
            loss = K.mean(K.sum(loss, axis=-1) / K.sum(weights))
        else:
            loss = K.mean(loss)

        return loss

    return d_loss


def weighted_categorical_crossentropy(weights, **kwargs):
    """A weighted version of Keras categorical_crossentropy

    :param: weights: Numpy array of shape (C,) where C is the number of classes

    Usage:
    ```
    weights = np.array([0.5, 2]) # Class 1 at 0.5, class 2 2x the normal weights
    exclude_ind = 0  # exclude background from computation
    loss = weighted_categorical_crossentropy(weights=weights,
    exclude=exclude_ind)
    model.compile(loss=loss, optimizer='adam')
    ```
    """

    weights = K.variable(weights)

    def loss(y_true, y_pred):
        # scale predictions so that the class probas of each sample sum to 1
        y_pred /= K.sum(y_pred, axis=-1, keepdims=True)
        # clip to prevent NaN's and Inf's
        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())
        # calc
        loss = y_true * K.log(y_pred) * weights
        loss = -K.sum(loss, -1)
        return loss

    return loss


def weighted_categorical_crossentropy_sigmoid(weights):
    """
    A weighted version of Keras categorical_crossentropy

    @:param: weights: numpy array of shape (C,) where C is the number of classes

    Use Case:
    ```
    weights = np.array([0.5, 2]) # Class 1 at 0.5, class 2 2x the normal weights
    loss = weighted_categorical_crossentropy(weights)
    model.compile(loss=loss,optimizer='adam')
    ```
    """

    weights = K.variable(weights)

    def loss(y_true, y_pred):
        # scale predictions so that the class probas of each sample sum to 1
        # y_pred /= K.sum(y_pred, axis=-1, keepdims=True)
        # clip to prevent NaN's and Inf's
        szp = _get_shape(y_pred)
        img_len = szp[1] * szp[2] * szp[3]

        y_true = K.reshape(y_true, (-1, img_len))
        y_pred = K.reshape(y_pred, (-1, img_len))
        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())

        # calc
        loss = (1 - y_true) * K.log(1 - y_pred) * weights[0] + y_true * K.log(y_pred) * weights[1]
        loss = -K.mean(loss)
        return loss

    return loss


def focal_loss(gamma=FOCAL_LOSS_GAMMA):
    """
    Focal loss as implemented by facebook

    formula is -(1 - pt)^gamma * log(pt),   pt = p if y=1
                                            pt = 1-p if y=0

    gamma = 3

    y_true values must be 0s and 1s

    @:param: weights: numpy array of shape (C,) where C is the number of classes
    """

    def f_loss(y_true, y_pred):
        y_true = K.flatten(y_true)
        y_pred = K.flatten(y_pred)

        y_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())

        # calculate pt
        # note that ~y_true = 1 - y_true
        pt = y_true * y_pred + (1 - y_true) * (1 - y_pred)

        loss_val = -K.mean(K.pow((1 - pt), gamma) * K.log(pt)) * 100

        return loss_val

    return f_loss


def wasserstein_disagreement_map(prediction, ground_truth, M):
    """
    Function to calculate the pixel-wise Wasserstein distance between the
    flattened pred_proba and the flattened labels (ground_truth) with respect
    to the distance matrix on the label space M.
    :param prediction: the logits after softmax
    :param ground_truth: segmentation ground_truth
    :param M: distance matrix on the label space
    :return: the pixelwise distance map (wass_dis_map)
    """
    # pixel-wise Wassertein distance (W) between flat_pred_proba and flat_labels
    # wrt the distance matrix on the label space M
    n_classes = K.int_shape(prediction)[-1]
    # unstack_labels = tf.unstack(ground_truth, axis=-1)
    ground_truth = tf.cast(ground_truth, dtype=tf.float64)
    # unstack_pred = tf.unstack(prediction, axis=-1)
    prediction = tf.cast(prediction, dtype=tf.float64)
    # logger.info("shape of M", M.shape, "unstacked labels", unstack_labels,
    #       "unstacked pred" ,unstack_pred)
    # W is a weighting sum of all pairwise correlations (pred_ci x labels_cj)
    pairwise_correlations = []
    for i in range(n_classes):
        for j in range(n_classes):
            pairwise_correlations.append(
                M[i, j] * tf.multiply(prediction[:, i], ground_truth[:, j])
            )
    wass_dis_map = tf.add_n(pairwise_correlations)
    return wass_dis_map


M_tree_4 = np.array(
    [[0.0, 1.0, 1.0, 1.0], [1.0, 0.0, 0.6, 0.5], [1.0, 0.6, 0.0, 0.7], [1.0, 0.5, 0.7, 0.0]],
    dtype=np.float64,
)


def generalised_wasserstein_dice_loss(y_true, y_predicted):
    """
    Function to calculate the Generalised Wasserstein Dice Loss defined in
    Fidon, L. et. al. (2017) Generalised Wasserstein Dice Score for Imbalanced
    Multi-class Segmentation using Holistic Convolutional Networks.
    MICCAI 2017 (BrainLes)
    :param prediction: the logits (before softmax)
    :param ground_truth: the segmentation ground_truth
    :param weight_map:
    :return: the loss
    """
    # apply softmax to pred scores
    n_classes = K.int_shape(y_predicted)[-1]

    ground_truth = tf.cast(tf.reshape(y_true, (-1, n_classes)), dtype=tf.int64)
    pred_proba = tf.cast(tf.reshape(y_predicted, (-1, n_classes)), dtype=tf.float64)

    # M = tf.cast(M, dtype=tf.float64)
    # compute disagreement map (delta)
    M = M_tree_4
    # logger.info("M shape is ", M.shape, pred_proba, one_hot)
    delta = wasserstein_disagreement_map(pred_proba, ground_truth, M)
    # compute generalisation of all error for multi-class seg
    all_error = tf.reduce_sum(delta)
    # compute generalisation of true positives for multi-class seg
    one_hot = tf.cast(ground_truth, dtype=tf.float64)
    true_pos = tf.reduce_sum(
        tf.multiply(tf.constant(M[0, :n_classes], dtype=tf.float64), one_hot), axis=1
    )
    true_pos = tf.reduce_sum(tf.multiply(true_pos, 1.0 - delta), axis=0)
    WGDL = 1.0 - (2.0 * true_pos) / (2.0 * true_pos + all_error)

    return tf.cast(WGDL, dtype=tf.float32)


def l2_loss(y_true, y_pred):
    """
    A basic L2 loss, which just compares the pixel values of two images or
    feature maps. The average L2 loss is returned, where the average is
    taken along the batch dimension.

    Args:
        y_true: The ground truth image or feature map.
        y_pred: The predicted image or feature map from the network.

    Returns:
       avg_l2_loss: A tensor, whose value is the average L2 loss.
    """
    square_diff = K.square(y_true - y_pred)
    ch_l2_loss = K.sum(square_diff, axis=(1, 2))
    avg_ch_l2 = K.mean(ch_l2_loss, axis=0)
    avg_l2_loss = K.mean(avg_ch_l2)
    return avg_l2_loss


class NaiveAdaRobLossComputer(Callback):
    """Handles adaptive robust class loss computer.

    Use `on_batch_begin` and `on_batch_end` to do things before/after the training batch.
    Note `on_batch_end` runs **before** validation when using `model.fit_generator`.
    Requires tf>=2.0 and eager mode execution.

    TODO:
        - Add adjustment
    """

    def __init__(self, criterion, n_groups, robust_step_size, stable=True):
        # TF2 requires __name__ attribute which is not set by default.
        self.__name__ = "NaiveAdaRobLossComputer"

        if not env.is_tf2():
            raise EnvironmentError(f"{self.__name__} only supported on tensorflow>=2.0")

        super().__init__()
        self.criterion = criterion

        self.n_groups = n_groups
        self.group_range = np.arange(self.n_groups, dtype=np.long)[..., np.newaxis]

        self.robust_step_size = robust_step_size
        logger.info(f"Using robust loss with inner step size {self.robust_step_size}")
        self.stable = stable

        # The following quantities are maintained/updated throughout training
        if self.stable:
            logger.info("Using numerically stabilized DRO algorithm")
            self.adv_probs_logits = np.zeros(self.n_groups)
        else:  # for debugging purposes
            logger.warning("Using original DRO algorithm")
            self.adv_probs = np.ones(self.n_groups) / self.n_groups

        self.training = None
        self.prev_training_state = None
        self._tmp = None  # Initialized in on_batch_begin (only used at train time)

    def loss(self, y_true, y_pred):
        # Get average classes losses (1D array - length C).
        # if not tf.executing_eagerly():
        #     raise RuntimeError(f"{self.__name__} does not support non-eager execution")

        group_losses = self.criterion(y_true, y_pred)  # Nx...xC
        if group_losses.ndim > 1:
            group_losses = tf.math.reduce_mean(group_losses, axis=range(0, group_losses.ndim - 1))
        # group_losses, group_counts = self.compute_group_avg(group_losses, y_true)

        # Compute overall loss.
        # The naive implementation does not do any sort of grouping.
        robust_loss = self.compute_robust_loss(group_losses)

        # Store for logging purposes.
        # Will be deleted after each batch.
        if self._tmp is not None:
            self._tmp["group_losses"] = group_losses.numpy()

        return robust_loss

    def compute_robust_loss(self, group_loss):
        # Requires eager execution
        if not tf.executing_eagerly():
            raise ValueError(f"{self.__name__} requires eager execution")

        assert (
            self.training is not None
        ), "`self.training` not initialized. Make sure this class is added as a callback"
        if self.training:
            # Update weighting if in training mode
            # This only works in eager mode.
            # TODO: Find solution for non-eager mode to speed up large-scale training
            adjusted_loss = group_loss if isinstance(group_loss, np.ndarray) else group_loss.numpy()
            logit_step = self.robust_step_size * adjusted_loss
            if self.stable:
                self.adv_probs_logits = self.adv_probs_logits + logit_step
            else:
                raise AssertionError("This branch has not been tested")
                # self.adv_probs = self.adv_probs * torch.exp(logit_step)
                # self.adv_probs = self.adv_probs / self.adv_probs.sum()

        if self.stable:
            adv_probs = (
                sps.softmax(self.adv_probs_logits)
                if isinstance(self.adv_probs_logits, np.ndarray)
                else K.softmax(self.adv_probs_logits, axis=-1)
            )
        else:
            adv_probs = self.adv_probs

        if isinstance(group_loss, np.ndarray):
            robust_loss = group_loss @ adv_probs
        else:
            robust_loss = K.sum(group_loss * adv_probs)
        return robust_loss

    def compute_group_avg(self, per_sample_losses, group_idx):
        """Compute average loss per group and counts of each group.

        Count is defined as the number of pixels corresponding to a group.
        Even for pixel-aggregate losses, such as dice, this is the default
        used.

        Args:
            per_sample_losses: Losses for each sample. Sample can either be batch
                or per-pixel.
        """
        group_count = tf.math.reduce_sum(group_idx, axis=range(1, group_idx.ndim - 1))
        if per_sample_losses.ndim == 2:  # NxC
            group_loss = tf.math.reduce_mean(per_sample_losses, axis=0)
        else:
            raise ValueError("`per_sample_losses` must have shape [N,C]")

        return group_loss, group_count

    def on_batch_begin(self, batch, logs=None):
        # Hacky way to turn on training before training batch in case it is off.
        self.training = True
        self._tmp = {}

    def on_batch_end(self, batch, logs=None):
        # Hacky way to turn off training when validation starts.
        # `fit_generator` executes `on_batch_end` before starting validation
        # This let's us turn off training mode during the validation period.
        self.training = False
        weighting = sps.softmax(self.adv_probs_logits)
        logs.update(
            {f"{self.__name__}/weights/class:{i}": weighting[i] for i in range(self.n_groups)}
        )
        group_losses = self._tmp["group_losses"]
        logs.update(
            {f"{self.__name__}/loss/class:{i}": group_losses[i] for i in range(self.n_groups)}
        )
        self._tmp = {}

    def on_test_begin(self, logs=None):
        self.prev_training_state = self.training
        self.training = False

    def on_test_end(self, logs=None):
        assert (
            self.prev_training_state is not None
        ), "`self.prev_training_state` should be set in `on_test_begin`"
        self.training = self.prev_training_state

    def __call__(self, y_true, y_pred):
        return self.loss(y_true, y_pred)



import logging

from keras import callbacks as kc
from keras.optimizers import Adam

from medsegpy.config import Config

from .lr_scheduler import step_decay
from .optimizer import AdamAccumulate

__all__ = ["build_lr_scheduler", "build_optimizer"]

logger = logging.getLogger(__name__)


def build_optimizer(config: Config):
    """Build optimizer from config.

    Currently supports :class:`Adam` or :class:`AdamAccumulate` optimizers.

    Args:
        config (Config): A config to read parameters from.

    Returns:
        A Keras-compatible optimizer.
    """
    if config.NUM_GRAD_STEPS == 1:
        optimizer = Adam(
            lr=config.INITIAL_LEARNING_RATE,
            beta_1=0.99,
            beta_2=0.995,
            epsilon=1e-8,
            decay=config.ADAM_DECAY,
            amsgrad=config.USE_AMSGRAD,
        )
    elif config.NUM_GRAD_STEPS > 1:
        logger.info("Accumulating gradient over {} steps".format(config.NUM_GRAD_STEPS))
        optimizer = AdamAccumulate(
            lr=config.INITIAL_LEARNING_RATE,
            beta_1=0.99,
            beta_2=0.995,
            epsilon=1e-8,
            decay=config.ADAM_DECAY,
            amsgrad=config.USE_AMSGRAD,
            accum_iters=config.NUM_GRAD_STEPS,
        )
    else:
        raise ValueError("config.NUM_GRAD_STEPS must be >= 1")

    return optimizer


def build_lr_scheduler(config: Config) -> kc.Callback:
    """Build learning rate scheduler.

    Supports "StepDecay" and "ReduceLROnPlateau"

        Args:
        config (Config): A config to read parameters from.

    Returns:
        :class:`keras.callback.LearningRateScheduler`

    Usage:
        >>> callbacks = []  # list of callbacks to be used sith `fit_generator`
        >>> scheduler = build_lr_scheduler(...)
        >>> callbacks.append(scheduler)
    """
    name = config.LR_SCHEDULER_NAME
    if name == "StepDecay":
        scheduler_func = step_decay(
            initial_lr=config.INITIAL_LEARNING_RATE,
            min_lr=config.MIN_LEARNING_RATE,
            drop_factor=config.DROP_FACTOR,
            drop_rate=config.DROP_RATE,
        )
        return kc.LearningRateScheduler(scheduler_func)
    elif name == "ReduceLROnPlateau":
        if config.LR_PATIENCE <= 0:
            raise ValueError("LR patience must be >= 0. Got {}".format(config.LR_PATIENCE))
        return kc.ReduceLROnPlateau(
            factor=config.DROP_FACTOR,
            patience=config.LR_PATIENCE,
            min_delta=config.LR_MIN_DELTA,
            cooldown=config.LR_COOLDOWN,
            min_lr=config.MIN_LEARNING_RATE,
        )
    else:
        raise ValueError("LR scheduler {} not supported".format(config.LR_SCHEDULER_NAME))



import json
import logging
import os
import re

from medsegpy.data.catalog import DatasetCatalog, MetadataCatalog

logger = logging.getLogger(__name__)

BASE_ANNOT_DIR = "/bmrNAS/people/arjun/data/qdess_knee_2020/annotations/v0.0.1"
BASE_IMAGE_DIR = "/dataNAS/people/arjun/data/qdess_knee_2020/image_files_total"
BASE_UNUSED_DIR = "/dataNAS/people/arjun/data/qdess_knee_2020/image_files_total"
BASE_UNLABELED_DIR = (
    "/dataNAS/people/arjun/data/qdess_knee_2020/unlabeled_outputs_dcm_fullanon_oct2021"
)

UNLABELED_SCANS = [
    "MTR_002",
    "MTR_003",
    "MTR_004",
    "MTR_007",
    "MTR_009",
    "MTR_011",
    "MTR_012",
    "MTR_014",
    "MTR_017",
    "MTR_021",
    "MTR_022",
    "MTR_024",
    "MTR_027",
    "MTR_029",
    "MTR_031",
    "MTR_032",
    "MTR_035",
    "MTR_036",
    "MTR_039",
    "MTR_041",
    "MTR_054",
    "MTR_055",
    "MTR_058",
    "MTR_059",
    "MTR_063",
    "MTR_064",
    "MTR_067",
    "MTR_068",
    "MTR_072",
    "MTR_073",
    "MTR_075",
    "MTR_081",
    "MTR_082",
    "MTR_085",
    "MTR_086",
    "MTR_089",
    "MTR_090",
    "MTR_091",
    "MTR_092",
    "MTR_102",
    "MTR_106",
    "MTR_108",
    "MTR_111",
    "MTR_114",
    "MTR_117",
    "MTR_119",
    "MTR_121",
    "MTR_123",
    "MTR_125",
    "MTR_128",
    "MTR_131",
    "MTR_132",
    "MTR_136",
    "MTR_137",
    "MTR_138",
    "MTR_153",
    "MTR_155",
    "MTR_157",
    "MTR_160",
    "MTR_162",
    "MTR_165",
    "MTR_166",
    "MTR_168",
    "MTR_169",
    "MTR_170",
    "MTR_181",
    "MTR_185",
    "MTR_195",
    "MTR_197",
    "MTR_231",
    "MTR_234",
    "MTR_242",
    "MTR_246",
    "MTR_247",
    "MTR_249",
]

UNUSED_SCANS = [
    "MTR_025",
    "MTR_042",
    "MTR_050",
    "MTR_060",
    "MTR_070",
    "MTR_076",
    "MTR_078",
    "MTR_088",
    "MTR_134",
    "MTR_141",
    "MTR_143",
    "MTR_147",
    "MTR_171",
    "MTR_180",
    "MTR_187",
    "MTR_202",
    "MTR_217",
    "MTR_226",
    "MTR_238",
    "MTR_239",
    "MTR_250",
]

_DATA_CATALOG = {
    "qdess_seg_train": os.path.join(BASE_ANNOT_DIR, "annotations/v0.0.1/train.json"),
    "qdess_seg_val": os.path.join(BASE_ANNOT_DIR, "annotations/v0.0.1/val.json"),
    "qdess_seg_test": os.path.join(BASE_ANNOT_DIR, "annotations/v0.0.1/test.json"),
    "qdess_seg_50_train": "qdess_data/50_perc_train/50_perc.json",
    "qdess_seg_25_train": "qdess_data/25_perc_train/25_perc.json",
    "qdess_seg_10_train": "qdess_data/10_perc_train/10_perc_split.json",
    "qdess_seg_5_train": "qdess_data/5_perc_train/5_perc_split.json",
    "qdess_125_perc_train": "qdess_data/percentage_jsons/125_perc.json",
    "qdess_150_perc_train": "qdess_data/percentage_jsons/150_perc.json",
    "qdess_175_perc_train": "qdess_data/percentage_jsons/175_perc.json",
    "qdess_200_perc_train": "qdess_data/percentage_jsons/200_perc.json",
    "qdess_unlabeled_all_train": "qdess_data/unlabeled_images/all_images.json",
    "qdess_seg_unused_train": "qdess_data/unlabeled_images/unused_orig_images.json",
}


QDESS_INPAINTING_CATEGORIES = [
    {"id": 0, "name": "inpainting_dummy_category", "abbrev": "inpainted", "color": [0, 0, 0]}
]

QDESS_SEGMENTATION_CATEGORIES = [
    {"id": 0, "name": "patellar cartilage", "abbrev": "pc", "color": [220, 20, 60]},
    {"id": 1, "name": "femoral cartilage", "abbrev": "fc", "color": [250, 170, 30]},
    {"id": (2, 3), "name": "tibial cartilage", "abbrev": "tc", "color": [0, 142, 0]},
    {"id": (4, 5), "name": "meniscus", "abbrev": "men", "color": [203, 92, 255]},
]


def load_2d_from_filepaths(filepaths: list, source_path: str, dataset_name: str = None):
    """
    Creates a list of dictionaries, where each dictionary contains information
    about each 2D image and corresponding ground truth segmentation in
    a particular directory (scan_root).

    The function assumes all 2D images are stored in the following format:

    %7d_V%02d-Aug%02d_%03d
    e.g. "9311328_V01-Aug04_072.im"

    Args:
        scan_root: The path of the directory containing the 2D images and
                    corresponding ground truth segmentations.
        total_num_slices: The total number of slices for this dataset.
        dataset_name: The name of the dataset.
    Returns:
        dataset_dicts: A list of dictionaries, described above in the
                        docstring.
    """
    FNAME_REGEX = "([\d]+)_V([\d]+)-Aug([\d]+)_([\d]+)"

    # Read appropriate JSON file for qDESS 2020 dataset
    split_type = dataset_name.split("_")[-1]
    annot_json_path = os.path.join(BASE_ANNOT_DIR, split_type + ".json")
    with open(annot_json_path, "r") as fp:
        annot_json = json.load(fp)
    volume_list = annot_json["images"]

    dataset_dicts = []
    for fp in filepaths:
        _, pid, time_point, _, slice_id, _ = tuple(re.split(FNAME_REGEX, fp))
        pid = int(pid)
        time_point = int(time_point)
        scan_id = f"{pid:07d}_V{time_point:02d}"

        # Iterate through image list of annotation JSON file to determine
        # number of slices in scan
        found_scan = False
        num_slices_in_scan = -1
        for volume in volume_list:
            if volume["msp_id"] == scan_id:
                found_scan = True
                num_slices_in_scan = volume["matrix_shape"][2]
        assert found_scan, f"Could not find scan with msp_id = {scan_id}"

        dataset_dicts.append(
            {
                "file_name": fp,
                "sem_seg_file": "{}.seg".format(os.path.splitext(fp)[0]),
                "scan_id": scan_id,
                "subject_id": pid,
                "time_point": time_point,
                "slice_id": int(slice_id),
                "scan_num_slices": num_slices_in_scan,
            }
        )

    num_scans = len({d["scan_id"] for d in dataset_dicts})
    num_subjects = len({d["subject_id"] for d in dataset_dicts})
    if dataset_name:
        logger.info("Loaded {} from {}".format(dataset_name, source_path))
    logger.info(
        "Loaded {} scans from {} subjects ({} slices)".format(
            num_scans, num_subjects, len(dataset_dicts)
        )
    )
    return dataset_dicts


def load_2d_from_dir(scan_root: str, dataset_name: str = None):
    files = sorted(os.listdir(scan_root))
    filepaths = [os.path.join(scan_root, f) for f in files if f.endswith(".im")]

    return load_2d_from_filepaths(filepaths, scan_root, dataset_name=dataset_name)


def load_2d_from_json(json_path: str, dataset_name: str = None):
    with open(json_path, "r") as fp:
        data_info = json.load(fp)
    base_dir = BASE_IMAGE_DIR
    scan_list = data_info["scans"]
    all_files = sorted(os.listdir(base_dir))
    all_images = [os.path.join(base_dir, f) for f in all_files if f.endswith(".im")]

    filepaths = []
    for img_path in all_images:
        img_name = img_path.split("/")[-1]
        scan_id = img_name.split("-")[0]
        if scan_id in scan_list:
            filepaths.append(img_path)

    return load_2d_from_filepaths(filepaths, json_path, dataset_name=dataset_name)


def load_qdess(json_file, dataset_name=None):
    dataset_names = [dataset_name]
    json_files = [json_file]

    dataset_dicts = []
    for json_path, data_name in zip(json_files, dataset_names):  # noqa: B007
        with open(json_path, "r") as f:
            data = json.load(f)

        images = data["images"]

        for img in images:
            if img["scan_id"] in UNLABELED_SCANS:
                base_dir = BASE_UNLABELED_DIR
            elif img["scan_id"] in UNUSED_SCANS:
                base_dir = BASE_UNUSED_DIR
            else:
                base_dir = BASE_IMAGE_DIR
            fp = os.path.join(base_dir, img["file_name"])
            if not os.path.isfile(fp):
                raise FileNotFoundError(f"File {fp} not found")
            dataset_dicts.append(
                {
                    "file_name": fp,
                    "sem_seg_file": fp,
                    "scan_id": img["scan_id"],
                    "subject_id": img["subject_id"],
                    "scan_spacing": img["voxel_spacing"],
                    "image_size": img["matrix_shape"],
                }
            )

    num_scans = len(dataset_dicts)
    num_subjects = len({d["subject_id"] for d in dataset_dicts})
    if dataset_name:
        logger.info("Loaded {} from {}".format(dataset_name, json_file))
    logger.info("Loaded {} scans from {} subjects".format(num_scans, num_subjects))
    return dataset_dicts


def register_qdess_dataset(scan_root: str, dataset_name: str):
    """
    Registers the qDESS knee MRI 2020 dataset and adds metadata for the
    dataset.

    The function will choose the correct list of categories based on
    "dataset_name". If the task is inpainting, the names of the datasets used
    for inpainting must include the word "inpainting".

    Args:
        scan_root: The path of the directory containing the 2D images and
                    corresponding ground truth segmentations.
        dataset_name: The name of the dataset.

    Returns:
        N/A
    """
    if scan_root.endswith(".json"):
        load_func = load_qdess
    else:
        load_func = load_2d_from_dir

    DatasetCatalog.register(dataset_name, lambda: load_func(scan_root, dataset_name))

    # Add Metadata about the qDESS MRI dataset
    MetadataCatalog.get(dataset_name).set(
        scan_root="dummy_root",
        category_ids={
            "segmentation": [x["id"] for x in QDESS_SEGMENTATION_CATEGORIES],
            "inpainting": [x["id"] for x in QDESS_INPAINTING_CATEGORIES],
        },
        category_abbreviations={
            "segmentation": [x["abbrev"] for x in QDESS_SEGMENTATION_CATEGORIES],
            "inpainting": [x["abbrev"] for x in QDESS_INPAINTING_CATEGORIES],
        },
        categories={
            "segmentation": [x["name"] for x in QDESS_SEGMENTATION_CATEGORIES],
            "inpainting": [x["name"] for x in QDESS_INPAINTING_CATEGORIES],
        },
        category_colors={
            "segmentation": [x["color"] for x in QDESS_SEGMENTATION_CATEGORIES],
            "inpainting": [x["color"] for x in QDESS_INPAINTING_CATEGORIES],
        },
        category_id_to_contiguous_id={
            "segmentation": {x["id"]: idx for idx, x in enumerate(QDESS_SEGMENTATION_CATEGORIES)},
            "inpainting": {x["id"]: idx for idx, x in enumerate(QDESS_INPAINTING_CATEGORIES)},
        },
        evaluator_type={"segmentation": "QDESSEvaluator", "inpainting": "SemSegEvaluator"},
    )


def register_all_qdess_datasets():
    """
    Registers all qDESS MRI datasets listed in _DATA_CATALOG.
    """
    for dataset_name, scan_root in _DATA_CATALOG.items():
        register_qdess_dataset(scan_root=scan_root, dataset_name=dataset_name)



import os
import shutil
import socket
import unittest

from fvcore.common.file_io import PathManager

from medsegpy.utils.cluster import _UNKNOWN, Cluster


class TestCluster(unittest.TestCase):
    _orig_env = None
    _settings_path = PathManager.get_local_path("mock_data://.settings")

    @classmethod
    def setUpClass(cls):
        cls._orig_env = dict(os.environ)
        os.environ["MEDSEGPY_SETTINGS"] = cls._settings_path

    @classmethod
    def tearDownClass(cls):
        os.environ.clear()
        os.environ.update(cls._orig_env)
        shutil.rmtree(cls._settings_path)

    def test_save_load_delete(self):
        hostname = socket.gethostname()
        cluster = Cluster("SAMPLE", [hostname])
        cluster.save()
        assert os.path.isfile(cluster.filepath()), cluster.filepath()

        cluster2 = cluster.from_config(cluster.filepath())
        assert all(cluster.__dict__[k] == cluster2.__dict__[k] for k in cluster.__dict__)

        cluster.delete()
        assert not os.path.isfile(cluster.filepath()), cluster.filepath()

    def test_cluster(self):
        hostname = socket.gethostname()
        cluster = Cluster("SAMPLE", [hostname])
        cluster.save()

        cluster2 = Cluster.cluster()
        assert all(cluster.__dict__[k] == cluster2.__dict__[k] for k in cluster.__dict__), (
            cluster.__dict__,
            cluster2.__dict__,
        )

    def test_all_clusters(self):
        hostname = socket.gethostname()
        cluster_names = []
        for i in range(4):
            name = f"SAMPLE-{i}"
            cluster_names.append(name)
            Cluster(name, [hostname]).save()
        all_clusters = Cluster.all_clusters()
        assert len(all_clusters) == 4
        shutil.rmtree(self._settings_path)

    def test_working_cluster(self):
        cluster = Cluster.working_cluster()

        Cluster.set_working_cluster(None)
        assert Cluster.working_cluster() == _UNKNOWN, Cluster.working_cluster()
        assert Cluster.working_cluster().data_dir == "./datasets"
        assert Cluster.working_cluster().results_dir == "./results"

        orig_env = dict(os.environ)
        os.environ["MEDSEGPY_DATASETS"] = "/path/to/datasets"
        assert Cluster.working_cluster().data_dir == "/path/to/datasets"
        os.environ["MEDSEGPY_RESULTS"] = "/path/to/results"
        assert Cluster.working_cluster().results_dir == "/path/to/results"
        os.environ.clear()
        os.environ.update(orig_env)

        Cluster.set_working_cluster(cluster)
        assert Cluster.working_cluster() == cluster, Cluster.working_cluster()


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import unittest

import h5py
import numpy as np
from fvcore.common.file_io import PathManager

from medsegpy.config import Config
from medsegpy.data import N5dDataLoader, PatchDataLoader

from .. import utils  # noqa


class TestPatchDataLoader(unittest.TestCase):
    IMG_SIZE = (10, 20, 30)
    NUM_CLASSES = 4
    FILE_PATH = "mock_data://temp_data/scan1.h5"

    @classmethod
    def setUpClass(cls):
        img = np.random.rand(*cls.IMG_SIZE).astype(np.float32)
        seg = (np.random.rand(*cls.IMG_SIZE, cls.NUM_CLASSES) >= 0.5).astype(np.uint8)

        file_path = PathManager.get_local_path(cls.FILE_PATH)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with h5py.File(file_path, "w") as f:
            f.create_dataset("volume", data=img)
            f.create_dataset("seg", data=seg)

    @classmethod
    def tearDownClass(cls):
        file_path = PathManager.get_local_path(cls.FILE_PATH)
        shutil.rmtree(os.path.dirname(file_path))

    def get_dataset_dicts(self):
        file_path = PathManager.get_local_path(self.FILE_PATH)
        return [
            {
                "file_name": file_path,
                "sem_seg_file": file_path,
                "scan_id": os.path.splitext(os.path.basename(file_path))[0],
                "image_size": self.IMG_SIZE,
            }
        ]

    def test_slicing(self):
        """Test selecting slices.

        Currently only verified for last dimension.
        """
        with h5py.File(PathManager.get_local_path(self.FILE_PATH), "r") as f:
            volume = f["volume"][:]
            mask = f["seg"][:]

        cfg = Config("")
        cfg.IMG_SIZE = (10, 20, 1)
        cfg.CATEGORIES = (0, 1, 3)
        dataset_dicts = self.get_dataset_dicts()

        # Simple slice
        data_loader = PatchDataLoader(cfg, dataset_dicts, is_test=False, shuffle=False)
        assert len(data_loader) == 30
        for i in range(30):
            img, seg = data_loader[i]
            img, seg = np.squeeze(img, axis=(0, 3)), np.squeeze(seg, axis=0)
            assert np.all(img == volume[:, :, i])
            assert np.all(seg == mask[:, :, i, cfg.CATEGORIES])


class TestN5dDataLoader(unittest.TestCase):
    IMG_SIZE = (10, 20, 30)
    NUM_CLASSES = 4
    FILE_PATH = "mock_data://temp_data/scan1.h5"

    @classmethod
    def setUpClass(cls):
        img = np.random.rand(*cls.IMG_SIZE).astype(np.float32)
        seg = (np.random.rand(*cls.IMG_SIZE, cls.NUM_CLASSES) >= 0.5).astype(np.uint8)

        file_path = PathManager.get_local_path(cls.FILE_PATH)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with h5py.File(file_path, "w") as f:
            f.create_dataset("volume", data=img)
            f.create_dataset("seg", data=seg)

    @classmethod
    def tearDownClass(cls):
        file_path = PathManager.get_local_path(cls.FILE_PATH)
        shutil.rmtree(os.path.dirname(file_path))

    def get_dataset_dicts(self):
        file_path = PathManager.get_local_path(self.FILE_PATH)
        return [
            {
                "file_name": file_path,
                "sem_seg_file": file_path,
                "scan_id": os.path.splitext(os.path.basename(file_path))[0],
                "image_size": self.IMG_SIZE,
            }
        ]

    def test_simple(self):
        """Test simple 2.5d configuration."""
        with h5py.File(PathManager.get_local_path(self.FILE_PATH), "r") as f:
            volume = f["volume"][:]
            mask = f["seg"][:]

        num_slices = 3
        window = num_slices // 2
        cfg = Config("")
        cfg.IMG_SIZE = (10, 20, num_slices)
        cfg.CATEGORIES = (0, 1, 3)
        dataset_dicts = self.get_dataset_dicts()

        # No padding
        # First and last slices will be skipped b/c no padding is used
        data_loader = N5dDataLoader(cfg, dataset_dicts, is_test=False, shuffle=False)
        assert len(data_loader) == 30 - 2 * (num_slices // 2)
        for idx, i in enumerate(range(1, 29)):
            img, seg = data_loader[idx]
            img, seg = np.squeeze(img, axis=0), np.squeeze(seg, axis=0)
            assert np.all(img == volume[:, :, i - window : i + window + 1])
            assert np.all(seg == mask[:, :, i, cfg.CATEGORIES])

    def test_padding(self):
        with h5py.File(PathManager.get_local_path(self.FILE_PATH), "r") as f:
            volume = f["volume"][:]
            mask = f["seg"][:]

        num_slices = 3
        window = num_slices // 2
        cfg = Config("")
        cfg.IMG_SIZE = (10, 20, num_slices)
        cfg.CATEGORIES = (0, 1, 3)
        dataset_dicts = self.get_dataset_dicts()

        pad_size = ((0, 0), (0, 0), (window, window))
        pad_mode = "edge"
        cfg.IMG_PAD_SIZE = (0, 0, window)
        cfg.IMG_PAD_MODE = pad_mode
        data_loader = N5dDataLoader(cfg, dataset_dicts, is_test=False, shuffle=False)
        _s_volume = np.pad(volume, pad_size, pad_mode)
        _s_mask = np.pad(mask, pad_size + ((0, 0),), pad_mode)
        assert len(data_loader) == 30
        for idx, i in enumerate(range(window, 30 + window)):
            img, seg = data_loader[idx]
            img, seg = np.squeeze(img, axis=0), np.squeeze(seg, axis=0)
            assert np.all(img == _s_volume[:, :, i - window : i + window + 1])
            assert np.all(seg == _s_mask[:, :, i, cfg.CATEGORIES])


if __name__ == "__main__":
    unittest.main()



"""Default engine activities.

Adapted from Facebook's detectron2.
https://github.com/facebookresearch/detectron2
"""
import argparse
import os
import warnings

import keras.backend as K
import tensorflow as tf
from fvcore.common.file_io import PathManager

from medsegpy import glob_constants
from medsegpy.utils import dl_utils, env
from medsegpy.utils.collect_env import collect_env_info
from medsegpy.utils.io_utils import format_exp_version
from medsegpy.utils.logger import setup_logger


def config_exists(experiment_dir: str):
    return (
        os.path.isfile(os.path.join(experiment_dir, "config.ini"))
        or os.path.isfile(os.path.join(experiment_dir, "config.yaml"))
        or os.path.isfile(os.path.join(experiment_dir, "config.yml"))
    )


def default_argument_parser():
    """
    Create a parser with some common arguments used by detectron2 users.

    Returns:
        argparse.ArgumentParser:
    """
    parser = argparse.ArgumentParser(description="MedSegPy Training")
    parser.add_argument("--config-file", default="", metavar="FILE", help="path to config file")
    parser.add_argument("--eval-only", action="store_true", help="perform evaluation only")
    parser.add_argument("--num-gpus", type=int, default=1, help="number of gpus")
    # parser.add_argument(
    #     "--overwrite", action="store_true", help="overwrite previous experiment"
    # )
    parser.add_argument("--debug", action="store_true", help="run in debug mode")

    # Add option to execute non-eagerly in tensorflow 2
    if env.is_tf2():
        parser.add_argument("--non-eagerly", action="store_true", help="run tensorflow non-eagerly")

    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )

    return parser


def default_setup(cfg, args):
    """
    Perform some basic common setups at the beginning of a job, including:

    1. Set up the medsegpy logger
    2. Set python random seed.
    3. Log basic information about environment, cmdline arguments, and config
    4. Backup the config to the output directory
    5. Version experiments

    Args:
        cfg (CfgNode): the full config to be used
        args (argparse.NameSpace): the command line arguments to be logged
    """
    # Make new experiment version if not evaluating
    # TODO: Add support to evaluate latest version (if found) when eval_only specified.
    cfg.OUTPUT_DIR = PathManager.get_local_path(cfg.OUTPUT_DIR)
    make_new_version = not (hasattr(args, "eval_only") and args.eval_only)
    if not make_new_version and not config_exists(cfg.OUTPUT_DIR):
        raise ValueError(
            f"Tried to evaluate on empty experiment directory. " f"{cfg.OUTPUT_DIR} does not exist."
        )

    if args.eval_only and config_exists(cfg.OUTPUT_DIR):
        pass
    elif args.debug:
        os.environ["MEDSEGPY_RUN_MODE"] = "debug"
        cfg.OUTPUT_DIR = (
            os.path.join(cfg.OUTPUT_DIR, "debug")
            if os.path.basename(cfg.OUTPUT_DIR).lower() != "debug"
            else cfg.OUTPUT_DIR
        )
    else:
        cfg.OUTPUT_DIR = format_exp_version(cfg.OUTPUT_DIR, new_version=make_new_version)

    # Setup cuda visible devices.
    num_gpus = args.num_gpus
    if num_gpus > 0:
        gpu_ids = dl_utils.get_available_gpus(num_gpus)
        gpu_ids_tf_str = ",".join([str(g_id) for g_id in gpu_ids])
        os.environ["CUDA_VISIBLE_DEVICES"] = gpu_ids_tf_str
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = ""
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

    # Generate seed (if not specified).
    if cfg.SEED == -1:
        cfg.SEED = env.generate_seed()

    # Set experiment name.
    if not cfg.EXP_NAME:
        cfg.EXP_NAME = default_exp_name(cfg)

    output_dir = cfg.OUTPUT_DIR
    PathManager.mkdirs(output_dir)

    setup_logger(output_dir, name="fvcore")
    logger = setup_logger(output_dir)

    # Set seed.
    env.seed_all_rng(cfg.SEED)

    # Log environment information
    logger.info("Environment info:\n" + collect_env_info())

    logger.info("Command line arguments: " + str(args))
    if hasattr(args, "config_file") and args.config_file != "":
        logger.info(
            "Contents of args.config_file={}:\n{}".format(
                args.config_file, PathManager.open(args.config_file, "r").read()
            )
        )

    logger.info("Running with full config:\n")

    cfg.summary()
    cfg.save_config()

    if cfg.SEED is not None:
        logger.info("Using seed {}".format(cfg.SEED))
        glob_constants.SEED = cfg.SEED

    # Set image format to be (N, dim1, dim2, dim3, channel).
    K.set_image_data_format("channels_last")

    # Non-eager execution in tf2
    if env.is_tf2() and args.non_eagerly:
        if env.tf_version() >= (2, 3):
            tf.config.run_functions_eagerly(False)
            logger.info("Disabling eager execution...")
        else:
            logger.warning(
                "Eager mode has not been disabled. May have to disable manually in the model"
            )


def default_exp_name(cfg):
    """Extracts default experiment name from the config.

    `cfg.EXP_NAME` if exists. If basename starts with "version" or "debug",
    take both parent directory name and version name to make experiment name
    (e.g. "my_exp/version_001").

    Returns:
        exp_name (str): The default convention for naming experiments.
    """
    exp_name = cfg.get("EXP_NAME")
    if not exp_name:
        warnings.warn("EXP_NAME not specified. Defaulting to basename...")
        basename = os.path.basename(cfg.OUTPUT_DIR)
        if basename.startswith("version") or basename.startswith("debug"):
            exp_name = f"{os.path.basename(os.path.dirname(cfg.OUTPUT_DIR))}/{basename}"
        else:
            exp_name = basename
    return exp_name



import copy
import logging
import os
import pickle
import warnings
from typing import Tuple

import tensorflow as tf
from keras import callbacks as kc
from keras.utils import plot_model

from medsegpy import config, solver
from medsegpy.data import build_loader, im_gens
from medsegpy.engine.callbacks import LossHistory, WandBLogger, lr_callback
from medsegpy.evaluation import build_evaluator, inference_on_dataset
from medsegpy.losses import build_loss, dice_loss
from medsegpy.modeling.meta_arch import build_model
from medsegpy.modeling.ssl_utils import SelfSupervisedInfo, load_specific_weights
from medsegpy.utils import dl_utils, env, io_utils

try:
    _SUPPORTS_DISTRIBUTED = True
    from tensorflow.distribute import MirroredStrategy
except ModuleNotFoundError:
    _SUPPORTS_DISTRIBUTED = False
    MirroredStrategy = None

logger = logging.getLogger(__name__)


class DefaultTrainer(object):
    """Default trainer for medical semantic segmentation."""

    def __init__(self, cfg: config.Config, run_eagerly=None, strategy=None):
        """
        Args:
            cfg (Config): An experiment config.
            run_eagerly (bool, optional): If `True`, runs eagerly.
                Only available in tensorflow>=2.0.
            strategy (tf.distribute.Strategy, optional): The strategy to use for training.
                Only available if `tf.distribute` package is available (tensorflow>=1.14).
        """
        self._cfg = cfg
        self._loss_history = None
        self._default_strategy = (
            tf.distribute.get_strategy() if _SUPPORTS_DISTRIBUTED else dl_utils.NoOpStrategy()
        )
        num_gpus = dl_utils.num_gpus()

        if not env.is_tf2() and run_eagerly is not None:
            warnings.warn("`run_eagerly` can only be specified in Tensorflow >2.0. " "Ignoring...")
            run_eagerly = None
        self._run_eagerly = run_eagerly

        if strategy is None:
            strategy = self._default_strategy
            if _SUPPORTS_DISTRIBUTED and num_gpus > 1:
                logger.info("Running multi gpu model")
                strategy = MirroredStrategy()
        self.strategy = strategy

        with self.strategy.scope():
            # Prepare for self-supervised learning, if needed
            SelfSupervisedInfo.clear()
            SelfSupervisedInfo.init_self_supervised(cfg)

            model = self.build_model(cfg)
            if cfg.PRETRAINED_WEIGHTS_PATH or cfg.INIT_WEIGHTS:
                self._init_model(model)

        plot_model(model, to_file=os.path.join(cfg.OUTPUT_DIR, "model.png"), show_shapes=True)
        model.summary(line_length=120, print_fn=lambda x: logger.info(x))
        model_json = model.to_json()
        model_json_save_path = os.path.join(cfg.OUTPUT_DIR, "model.json")
        with open(model_json_save_path, "w") as json_file:
            json_file.write(model_json)

        # Replicate model on multiple gpus when tensorflow.distribute module not available.
        # Note this does not solve issue of having too large of a model
        if not _SUPPORTS_DISTRIBUTED and num_gpus > 1:
            logger.info("Running multi gpu model")
            model = dl_utils.ModelMGPU(model, gpus=num_gpus)

        self._train_loader, self._val_loader = self._build_data_loaders(cfg)
        self._model = model

    def train(self):
        """Train model specified by config.

        Do not call this under a strategy scope. Instead, set `self.strategy`.
        """
        cfg = self._cfg

        with self.strategy.scope():
            self._train_model()
            # After training, remove preloaded data
            self._train_loader.clear_cached_data()
            self._val_loader.clear_cached_data()

        if cfg.TEST_DATASET:
            # Specialized strategies are not currently supported for testing.
            if _SUPPORTS_DISTRIBUTED and not isinstance(
                self.strategy, (dl_utils.NoOpStrategy, type(self._default_strategy))
            ):
                logger.error(
                    f"Strategy '{type(self.strategy).__name__}' not currently "
                    f"supported for testing. "
                    f"Please run testing separately on a single gpu."
                )
                return {}
            return self.test(cfg, self._model)
        else:
            return {}

    def _init_model(self, model):
        """Initialize model with weights and apply any freezing necessary."""
        cfg = self._cfg
        if cfg.PRETRAINED_WEIGHTS_PATH:
            load_specific_weights(model, cfg, debug=True)
        else:
            if os.path.isdir(cfg.INIT_WEIGHTS):
                weight_file = dl_utils.get_weights(cfg.INIT_WEIGHTS)
            else:
                weight_file = cfg.INIT_WEIGHTS
            logger.info("Loading weights from {}".format(weight_file))
            model.load_weights(weight_file)
        frozen_layers = cfg.FREEZE_LAYERS
        if frozen_layers:
            fl = range(frozen_layers[0], frozen_layers[1])
            logger.info("Freezing layers [{}, {})".format(fl.start, fl.stop))
            for i in fl:
                model.layers[i].trainable = False

    def build_callbacks(self):
        cfg = self._cfg
        output_dir = cfg.OUTPUT_DIR
        callbacks = []

        if cfg.LR_SCHEDULER_NAME:
            callbacks.append(solver.build_lr_scheduler(cfg))
        if cfg.USE_EARLY_STOPPING:
            callbacks.append(
                kc.EarlyStopping(
                    monitor=cfg.EARLY_STOPPING_CRITERION,
                    min_delta=cfg.EARLY_STOPPING_MIN_DELTA,
                    patience=cfg.EARLY_STOPPING_PATIENCE,
                )
            )

        self._loss_history = LossHistory()

        tb_kwargs = dict(update_freq="batch") if env.is_tf2() else {}
        callbacks.extend(
            [
                kc.ModelCheckpoint(
                    os.path.join(output_dir, "weights.{epoch:03d}-{val_loss:.4f}.h5"),
                    save_best_only=True,
                    save_weights_only=True,
                ),
                kc.TensorBoard(output_dir, write_grads=False, write_images=False, **tb_kwargs),
                WandBLogger() if env.supports_wandb() else None,
                kc.CSVLogger(os.path.join(output_dir, "metrics.log")),
                self._loss_history,
            ]
        )
        callbacks = [x for x in callbacks if x is not None]

        return callbacks

    def build_loss(self):
        """Builds loss function used with ``model.compile(loss=...)``."""
        return build_loss(self._cfg)

    def _train_model(self):
        """Train model.

        If multi-gpu training and distributed training is supported (tensorflow>=1.15),
        call this function with the appropriate strategy scope::

            with self.strategy.scope():
                self._train_model()
        """
        cfg = self._cfg
        n_epochs = cfg.N_EPOCHS
        num_workers = cfg.NUM_WORKERS
        output_dir = cfg.OUTPUT_DIR

        model = self._model

        # TODO: Add more options for metrics.
        optimizer = solver.build_optimizer(cfg)
        loss_func = self.build_loss()
        metrics = [lr_callback(optimizer), dice_loss]

        callbacks = self.build_callbacks()
        if isinstance(loss_func, kc.Callback):
            callbacks.insert(0, loss_func)
            metrics.append(loss_func.criterion)

        model.compile(optimizer=optimizer, loss=loss_func, metrics=metrics)
        if env.is_tf2():
            run_eagerly = tf.executing_eagerly() if self._run_eagerly is None else self._run_eagerly
            model.run_eagerly = run_eagerly

        train_loader, val_loader = self._train_loader, self._val_loader
        use_multiprocessing = num_workers > 1

        # Start training
        model.fit_generator(
            train_loader,
            epochs=n_epochs,
            validation_data=val_loader,
            callbacks=callbacks,
            workers=num_workers,
            use_multiprocessing=use_multiprocessing,
            verbose=1,
            shuffle=False,
        )

        # Save optimizer state
        io_utils.save_optimizer(model.optimizer, output_dir)

        # Save files to write as output
        # TODO: refactor to save dataframe.
        hist_cb = self._loss_history
        data = [hist_cb.epoch, hist_cb.losses, hist_cb.val_losses]
        pik_data_path = os.path.join(output_dir, "pik_data.dat")
        with open(pik_data_path, "wb") as f:
            pickle.dump(data, f)

    @classmethod
    def test(cls, cfg: config.Config, model):
        logger.info("Beginning testing...")
        cfg = copy.deepcopy(cfg)  # will be modified below.
        cfg.change_to_test()

        weights = cfg.TEST_WEIGHT_PATH
        if not cfg.TEST_WEIGHT_PATH:
            weights = dl_utils.get_weights(cfg.OUTPUT_DIR)
            logger.info("Best weights: {}".format(weights))
        model.load_weights(weights)

        test_dataset = cfg.TEST_DATASET
        test_gen = cls.build_test_data_loader(cfg)
        evaluator = build_evaluator(test_dataset, cfg, save_raw_data=False)
        test_results = inference_on_dataset(model, test_gen, evaluator)
        # After testing, remove pre-loaded data
        test_gen.clear_cached_data()
        return test_results

    @classmethod
    def build_model(cls, cfg):
        try:
            return build_model(cfg)
        except KeyError:
            # TODO (TF2.X)
            if env.is_tf2():
                raise ValueError(
                    "`get_model` not currently supported for tf2. "
                    "We are working on backwards compatibility"
                )
            from medsegpy.modeling import get_model

            return get_model(cfg)

    def _build_data_loaders(self, cfg) -> Tuple[im_gens.Generator, im_gens.Generator]:
        """Builds train and val data loaders."""
        train_loader = build_loader(
            cfg,
            dataset_names=cfg.TRAIN_DATASET,
            batch_size=cfg.TRAIN_BATCH_SIZE,
            drop_last=True,
            is_test=False,
            shuffle=True,
        )
        val_loader = build_loader(
            cfg,
            dataset_names=cfg.VAL_DATASET,
            batch_size=cfg.VALID_BATCH_SIZE,
            drop_last=True,
            is_test=False,
            shuffle=False,
        )
        return train_loader, val_loader

    @classmethod
    def build_test_data_loader(cls, cfg):
        return build_loader(
            cfg,
            dataset_names=cfg.TEST_DATASET,
            batch_size=cfg.TEST_BATCH_SIZE,
            drop_last=False,
            is_test=True,
            shuffle=False,
        )



import os
import sys

import keras
import numpy as np
import tensorflow as tf
from keras import backend as K
from tabulate import tabulate


def _pretty_print_keras_config(env_str):
    # TODO (TF2.X)
    if not hasattr(K, "_config"):
        return "\n"

    cfg = K._config

    for k, v in cfg.items():
        env_str += "{}: {}\n".format(k, v)
    return env_str


def collect_env_info():
    data = []
    data.append(("sys.platform", sys.platform))
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    data.append(("Keras", keras.__version__))
    data.append(("Tensorflow", tf.__version__))

    try:
        import PIL

        data.append(("Pillow", PIL.__version__))
    except ImportError:
        pass

    try:
        import cv2

        data.append(("cv2", cv2.__version__))
    except ImportError:
        pass
    env_str = tabulate(data) + "\n"
    env_str += _pretty_print_keras_config(env_str)
    env_str += "CUDA Devices: {}".format(os.environ["CUDA_VISIBLE_DEVICES"])
    return env_str
